# from importlib import metadata

from .image import Image
from .collection import Collection
from . import interpolate

MODEL_NAME = 'SIMS'

# __version__ = metadata.version(__package__ or __name__)



# import pprint

import ee
import pytest

import openet.sims.data as data
import openet.sims.model as model
import openet.sims.utils as utils

# DATE = '2017-07-16'
YEAR = 2017
DOY = 197


# CGM - Should these be test fixtures instead?
#   I'm not sure how to make them fixtures and allow input parameters
# CGM - Setting default crop_type_source to 2017 image to simplify testing
#   but the default in the Model class init is the full collection
def default_model_args(
        year=YEAR,
        doy=DOY,
        crop_type_remap='CDL',
        crop_type_source=f'USDA/NASS/CDL/{YEAR}',
        crop_type_kc_flag=False,
        crop_type_annual_skip_flag=False,
        mask_non_ag_flag=False,
        water_kc_flag=True,
        reflectance_type='SR',
        ):
    return {
        'year': year,
        'doy': doy,
        'crop_type_source': crop_type_source,
        'crop_type_remap': crop_type_remap,
        'crop_type_kc_flag': crop_type_kc_flag,
        'crop_type_annual_skip_flag': crop_type_annual_skip_flag,
        'mask_non_ag_flag': mask_non_ag_flag,
        'water_kc_flag': water_kc_flag,
        'reflectance_type': reflectance_type,
    }


def default_model_obj(
        year=YEAR,
        doy=DOY,
        crop_type_remap='CDL',
        crop_type_source=f'USDA/NASS/CDL/{YEAR}',
        crop_type_kc_flag=False,
        crop_type_annual_skip_flag=False,
        mask_non_ag_flag=False,
        water_kc_flag=True,
        reflectance_type='SR',
        ):
    return model.Model(**default_model_args(
        year=ee.Number(year),
        doy=ee.Number(doy),
        crop_type_source=crop_type_source,
        crop_type_remap=crop_type_remap,
        crop_type_kc_flag=crop_type_kc_flag,
        crop_type_annual_skip_flag=crop_type_annual_skip_flag,
        mask_non_ag_flag=mask_non_ag_flag,
        water_kc_flag=water_kc_flag,
        reflectance_type=reflectance_type,
    ))


def test_crop_data_image():
    output = utils.constant_image_value(model.crop_data_image(
        param_name='crop_class',
        crop_type=ee.Image.constant(9).rename(['crop_class']),
        crop_data={9: {'crop_class': 10}},
    ))
    assert output['crop_class'] == 10


def test_crop_data_image_int_scalar():
    # Test that floating point values are scaled to int before remapping
    output = utils.constant_image_value(model.crop_data_image(
        param_name='m_l',
        crop_type=ee.Image.constant(9).rename(['m_l']),
        crop_data={9: {'m_l': 0.01}},
    ))
    assert output['m_l'] == 0.01


def test_crop_data_image_default_value():
    output = utils.constant_image_value(model.crop_data_image(
        param_name='crop_class',
        crop_type=ee.Image.constant(-999).rename(['crop_class']),
        crop_data={9: {'crop_class': 10}},
        default_value=100,
    ))
    assert output['crop_class'] == 100


def test_crop_data_image_default_nodata():
    output = utils.constant_image_value(model.crop_data_image(
        param_name='crop_class',
        crop_type=ee.Image.constant(-999).rename(['crop_class']),
        crop_data={9: {'crop_class': 10}},
    ))
    assert output['crop_class'] is None


def test_Model_init_default_parameters():
    m = default_model_obj()
    assert m.crop_type_source == f'USDA/NASS/CDL/{YEAR}'
    assert m.crop_type_remap == 'CDL'
    assert m.reflectance_type == 'SR'


@pytest.mark.parametrize(
    'parameter', ['m_l', 'h_max', 'fr_mid', 'fr_end', 'ls_start', 'ls_stop'])
def test_Model_init_crop_data_images(parameter):
    output = utils.getinfo(getattr(default_model_obj(), parameter))
    assert output['bands'][0]['id'] == parameter


@pytest.mark.parametrize(
    'year, expected',
    [
        [2007, 2008],
        [2008, 2008],
        [2016, 2016],
        [2023, 2023],
        # TODO: Make this test dynamic since it will fail in 2025 when the 2024 CDL is released
        [2024, 2023],
    ]
)
def test_Model_crop_type_source_cdl_collection(year, expected):
    """Test that the CDL collection is filtered to a single year and is limited
    to years with data (2008-2022 as of 6/7/2023)
    """
    output = utils.getinfo(default_model_obj(
        crop_type_source='USDA/NASS/CDL', year=ee.Number(year)).crop_type)
    assert output['properties']['id'] == f'USDA/NASS/CDL/{expected}'


def test_Model_crop_type_source_cdl_image():
    output = utils.getinfo(default_model_obj(crop_type_source='USDA/NASS/CDL/2008').crop_type)
    assert output['properties']['id'] == 'USDA/NASS/CDL/2008'


def test_Model_crop_type_source_cdl_image_exception():
    """Requesting a CDL image that doesn't exist should raise an EE exception"""
    with pytest.raises(Exception):
        # Intentionally using .getInfo() since utils.getinfo() might catch the exception
        default_model_obj(crop_type_source='USDA/NASS/CDL/2099').crop_type.getInfo()


@pytest.mark.parametrize(
    'crop_type_source',
    [
        'projects/openet/assets/crop_type/v2023a',
        'projects/openet/assets/crop_type/v2021a',
        # 'projects/openet/crop_type/v2021a',
        # 'projects/earthengine-legacy/assets/projects/openet/crop_type/v2021a',
    ]
)
def test_Model_crop_type_source_openet_crop_type(crop_type_source):
    output = utils.getinfo(default_model_obj(crop_type_source=crop_type_source).crop_type)
    expected = crop_type_source.replace('projects/earthengine-legacy/assets/', '')
    assert output['properties']['id'] == expected


def test_Model_crop_type_source_exception():
    with pytest.raises(ValueError):
        utils.getinfo(default_model_obj(crop_type_source='FOO'))
        # utils.getinfo(default_model_obj(crop_type_source='FOO').crop_type)


def test_Model_crop_type_constant_value():
    output = utils.constant_image_value(default_model_obj(crop_type_source=10).crop_type)
    assert output['crop_type'] == 10


def test_Model_crop_data_dictionary():
    assert default_model_obj(crop_type_remap='CDL').crop_data


def test_Model_crop_data_remap_exception():
    with pytest.raises(ValueError):
        utils.getinfo(default_model_obj(crop_type_remap='FOO'))
        # utils.getinfo(default_model_obj(crop_type_remap='FOO').crop_data)


@pytest.mark.parametrize('crop_type, parameter', [[1, 'h_max'], [1, 'm_l']])
def test_Model_crop_data_image(crop_type, parameter):
    output = utils.constant_image_value(
        getattr(default_model_obj(crop_type_source=crop_type), parameter)
    )
    assert output[parameter] == data.cdl[crop_type][parameter]


@pytest.mark.parametrize(
    'crop_type, expected',
    [
        [1, 1],
        [69, 2],
        [66, 3],
        [3, 5],    # Rice was switched to class 5 instead of 1
        [61, 6],   # Fallow was switched to class 6 instead of 1
        [176, 7],  # Grass/pasture was switched to class 7 instead of 1
    ]
)
def test_Model_crop_class_constant_value(crop_type, expected):
    output = utils.constant_image_value(
        default_model_obj(crop_type_source=crop_type, crop_type_remap='CDL').crop_class
    )
    assert output['crop_class'] == expected


@pytest.mark.parametrize(
    'ndvi, expected',
    [
        [-0.2, 0.0],  # Clamped
        [-0.1, 0.0],  # Clamped
        [0.0, 0.0],   # Clamped
        [0.1, 0.0],   # Clamped
        [0.2, 0.072],
        [0.5, 0.45],
        [0.7, 0.702],
        [0.8, 0.828],
        [0.95, 1.0],  # Clamped
    ]
)
def test_Model_fc_reflectance_type_sr(ndvi, expected, tol=0.0001):
    m = default_model_obj(reflectance_type='SR')
    output = utils.constant_image_value(m.fc(ndvi=ee.Image.constant(ndvi)))
    assert abs(output['fc'] - expected) <= tol


@pytest.mark.parametrize(
    'ndvi, expected',
    [
        [-0.2, 0.0],   # Clamped
        [-0.1, 0.0],   # Clamped
        [0.0, 0.0],    # Clamped
        [0.1, 0.0075],
        [0.14, 0.0661],   # Break value in rice Kc
        [0.2, 0.154],
        [0.35, 0.37375],  # Break value in fallow Kc
        [0.5, 0.5935],
        [0.7, 0.8865],
        [0.778, 1.0],  # Clamped
        [0.95, 1.0],   # Clamped
    ]
)
def test_Model_fc_reflectance_type_toa(ndvi, expected, tol=0.0001):
    m = default_model_obj(reflectance_type='TOA')
    output = utils.constant_image_value(m.fc(ndvi=ee.Image.constant(ndvi)))
    assert abs(output['fc'] - expected) <= tol


def test_Model_fc_reflectance_type_exception():
    with pytest.raises(Exception):
        utils.getinfo(default_model_obj(reflectance_type='FOO').fc(ndvi=ee.Image.constant(0.2)))


@pytest.mark.parametrize(
    'ndvi, expected',
    [
        [-0.2, 0.0],  # -0.05 without >= 0 clamping
        [-0.1, 0.075],
        [0.0, 0.2],
        [0.2, 0.45],
        [0.5, 0.825],
        [0.7, 1.075],
        [0.8, 1.2],
    ]
)
def test_Model_kc_generic_constant_value(ndvi, expected, tol=0.0001):
    m = default_model_obj(crop_type_source=1, crop_type_kc_flag=False)
    output = utils.constant_image_value(m.kc_generic(ndvi=ee.Image.constant(ndvi)))
    assert abs(output['kc'] - expected) <= tol


@pytest.mark.parametrize(
    'fc, expected',
    [
        [0.0, 0.15],
        [0.8, 0.9684],
    ]
)
def test_Model_kc_row_crop_constant_value(fc, expected, tol=0.0001):
    m = default_model_obj(crop_type_source=1, crop_type_kc_flag=False)
    output = utils.constant_image_value(m.kc_row_crop(fc=ee.Image.constant(fc)))
    assert abs(output['kc'] - expected) <= tol


def test_Model_kc_tree_constant_value(fc=0.8, expected=0.8*1.48+0.007, tol=0.0001):
    m = default_model_obj(crop_type_source=66, crop_type_kc_flag=False)
    output = utils.constant_image_value(m.kc_tree(fc=ee.Image.constant(fc)))
    assert abs(output['kc'] - expected) <= tol


@pytest.mark.parametrize(
    'ndvi, fc, expected',
    [
        [-0.1, 0.0, 1.05],
        [0.14, 0.0661, 1.05],
        # An NDVI in the range [0.14, 0.142857] will be clamped to 0 in fc(),
        # but is above the 0.14 threshold in kc() so it is not set to 1.05.
        [0.142, 0, 0.15],
        [0.143, 0.00018, 0.1503],
        [0.5, 0.45, 0.6855],
        [0.8, 0.828, 0.9860],
    ]
)
def test_Model_kc_rice_constant_value(ndvi, fc, expected, tol=0.0001):
    m = default_model_obj(crop_type_source=3, crop_type_kc_flag=False)
    output = utils.constant_image_value(
        m.kc_rice(fc=ee.Image.constant(fc), ndvi=ee.Image.constant(ndvi))
    )
    assert abs(output['kc'] - expected) <= tol


@pytest.mark.parametrize(
    'ndvi, fc, expected',
    [
        [-0.1, 0.0, 0.01],
        [0.1, 0.0075, 0.01],
        [0.2, 0.154, 0.154],
        [0.35, 0.37375, 0.37375],
        [0.351, 0.37375, 0.6084],
        [0.5, 0.5935, 0.8156],
        [0.8, 1.0, 1.0776],
    ]
)
def test_Model_kc_fallow_constant_value(ndvi, fc, expected, tol=0.0001):
    m = default_model_obj(crop_type_source=61, crop_type_kc_flag=False)
    output = utils.constant_image_value(
        m.kc_fallow(fc=ee.Image.constant(fc), ndvi=ee.Image.constant(ndvi))
    )
    assert abs(output['kc'] - expected) <= tol


@pytest.mark.parametrize(
    'ndvi, fc, expected',
    [
        [-0.1, 0.0, 0.01],
        [0.1, 0.0075, 0.01],
        [0.2, 0.154, 0.154],
        [0.35, 0.37375, 0.37375],
        [0.351, 0.37375, 0.6084],
        [0.5, 0.5935, 0.8156],
        [0.8, 1.0, 1.0776],
    ]
)
def test_Model_kc_grass_pasture_constant_value(ndvi, fc, expected, tol=0.0001):
    m = default_model_obj(crop_type_source=176, crop_type_kc_flag=False)
    output = utils.constant_image_value(
        m.kc_grass_pasture(fc=ee.Image.constant(fc), ndvi=ee.Image.constant(ndvi))
    )
    assert abs(output['kc'] - expected) <= tol


@pytest.mark.parametrize(
    'fc, expected',
    [
        # [-0.1, 0.0],
        [0.0, 0.0],
        [0.1, 0.1668],
        [0.2, 0.3591],
        [0.3, 0.5229],
        [0.4, 0.6521],
        [0.45, 0.7051],  # NDVI == 0.5
        [0.5, 0.7517],
        [0.6, 0.8284],
        [0.7, 0.8879],
        [0.8, 0.9283],
        [0.9, 0.9655],
        [1.0, 1.0],
        [1.1, 1.0],  # Check clamping
    ]
)
def test_Model_kd_row_crop_constant_value(fc, expected, tol=0.0001):
    # m_l for crop_type 1 == 2.0
    m = default_model_obj(crop_type_source=1, crop_type_kc_flag=False)
    output = utils.constant_image_value(m._kd_row_crop(fc=ee.Image.constant(fc)))
    assert abs(output['kd'] - expected) <= tol


@pytest.mark.parametrize(
    'fc, expected',
    [
        # [-0.1, 0.0],
        [0.0, 0.0],
        [0.45, 0.675],
        [0.5, 0.75],
        [0.6, 0.8434],  # 0.6 ** (1/3)
        # [0.7, 0.8879],  # 0.7 ** (1/3)
        [0.8, 0.9283],  # 0.8 ** (1/3)
        # [0.9, 0.9655],  # 0.9 ** (1/3)
        [1.0, 1.0],  # 1.0 ** (1/3)
        [1.1, 1.0],  # 1.0 ** (1/3)
    ]
)
def test_Model_kd_vine_constant_value(fc, expected, tol=0.0001):
    m = default_model_obj(crop_type_source=69, crop_type_kc_flag=False)
    output = utils.constant_image_value(m._kd_vine(fc=ee.Image.constant(fc)))
    assert abs(output['kd'] - expected) <= tol


@pytest.mark.parametrize(
    'fc, expected',
    [
        # [-0.1, 0.4096],
        [0.0, 0.0],
        [0.4, 0.7368],
        [0.45, 0.7663],
        [0.5, 0.7937],
        [0.6, 0.8801],
        [1.0, 1.0],
        [1.1, 1.0],
    ]
)
def test_Model_kd_tree_constant_value(fc, expected, tol=0.0001):
    m = default_model_obj(crop_type_source=66, crop_type_kc_flag=False)
    output = utils.constant_image_value(m._kd_tree(fc=ee.Image.constant(fc)))
    assert abs(output['kd'] - expected) <= tol


@pytest.mark.parametrize(
    'kd, doy, h_max, expected',
    [
        [1, 250, 3, min(3 * 0.1 + 1, 1.2) * 0.95],  # 1.14
        [1, 270, 3, min(3 * 0.1 + 1, 1.2) * 0.95],  # 1.14
        [1, 285, 3, min(3 * 0.1 + 1, 1.2) * 0.85],  # 1.02
        [1, 320, 3, min(3 * 0.1 + 1, 1.2) * 0.75],  # 0.9
        [1, 285, 1, min(1 * 0.1 + 1, 1.2) * 0.85],  # 0.935
        [0.5, 285, 3, 0.5 * (1.2 * 0.85 - 0.15) + 0.15],  # 0.585
    ]
)
def test_Model_kcb_constant_value(kd, doy, h_max, expected, tol=0.0001):
    m = default_model_obj(crop_type_source=66, doy=doy, crop_type_kc_flag=False)
    m.h_max = ee.Image.constant(h_max)
    m.fr_mid = ee.Image.constant(0.95)
    m.fr_end = ee.Image.constant(0.75)
    m.ls_start = ee.Image.constant(270)
    m.ls_stop = ee.Image.constant(300)
    output = utils.constant_image_value(m._kcb(kd=ee.Image.constant(kd)))
    assert abs(output['kcb'] - expected) <= tol


@pytest.mark.parametrize('crop_type', [0, 1, 69, 66, 3, 61])
def test_Model_kc_crop_class_constant_value(crop_type):
    # Check that a number is returned for all crop classes
    m = default_model_obj(crop_type_source=crop_type, crop_type_kc_flag=False)
    output = utils.constant_image_value(m.kc(ndvi=ee.Image.constant(0.5)))
    assert output['kc'] is not None


def test_Model_kc_crop_class_2_clamping():
    """Check that Kc for crop class 2 is clamped to 1.1"""
    m = default_model_obj(crop_type_source=78, crop_type_kc_flag=False)
    output = utils.constant_image_value(m.kc(ndvi=ee.Image.constant(0.85)))
    assert output['kc'] == 1.1


@pytest.mark.parametrize(
    'mask_non_ag_flag, expected',
    [
        [False, 0.825],
        [True, None],
    ]
)
def test_Model_kc_mask_non_ag_flag(mask_non_ag_flag, expected):
    m = default_model_obj(crop_type_source=0, mask_non_ag_flag=mask_non_ag_flag)
    output = utils.constant_image_value(m.kc(ndvi=ee.Image.constant(0.5)))
    assert output['kc'] == expected


@pytest.mark.parametrize(
    'crop_type_kc_flag, crop_type_annual_skip_flag',
    [
        [False, False],
        [True, False],
        [False, True],  # Use custom coefficients only for this condition
        [True, True],
    ]
)
def test_Model_kc_crop_type_kc_class_1(crop_type_kc_flag, crop_type_annual_skip_flag):
    """"Test that custom crop coefficients for annual crops are only used if
    the crop_type_kc_flag is True and the crop_type_annual_skip_flag is False.
    Otherwise, just use the generic row crop kc function.
    """
    m = default_model_obj(
        crop_type_source=1, crop_type_kc_flag=crop_type_kc_flag,
        crop_type_annual_skip_flag=crop_type_annual_skip_flag
    )
    output = utils.constant_image_value(m.kc(ndvi=ee.Image.constant(0.5)))
    if crop_type_kc_flag and not crop_type_annual_skip_flag:
        expected = utils.constant_image_value(m._kcb(m._kd_row_crop(fc=ee.Image.constant(0.45))))
        assert output['kc'] == expected['kcb']
    else:
        expected = utils.constant_image_value(m.kc_row_crop(fc=ee.Image.constant(0.45)))
        assert output['kc'] == expected['kc']


@pytest.mark.parametrize(
    'crop_type_kc_flag',
    [
        False,
        True,
    ]
)
def test_Model_kc_crop_type_kc_class_2(crop_type_kc_flag):
    """Check that vines are computed the same way """
    m = default_model_obj(crop_type_source=69, crop_type_kc_flag=crop_type_kc_flag)
    output = utils.constant_image_value(m.kc(ndvi=ee.Image.constant(0.5)))
    expected = utils.constant_image_value(m._kcb(m._kd_vine(fc=ee.Image.constant(0.45))))
    assert output['kc'] == expected['kcb']


@pytest.mark.parametrize(
    'crop_type_kc_flag',
    [
        False,
        True,
    ]
)
def test_Model_kc_crop_type_kc_class_3(crop_type_kc_flag):
    m = default_model_obj(crop_type_source=66, crop_type_kc_flag=crop_type_kc_flag)
    output = utils.constant_image_value(m.kc(ndvi=ee.Image.constant(0.5)))
    if crop_type_kc_flag:
        expected = utils.constant_image_value(m._kcb(m._kd_tree(fc=ee.Image.constant(0.45))))
        assert output['kc'] == expected['kcb']
    else:
        expected = utils.constant_image_value(m.kc_tree(fc=ee.Image.constant(0.45)))
        assert output['kc'] == expected['kc']


@pytest.mark.parametrize(
    'crop_type_source, expected',
    [
        # Wine grapes don't hit clamp value
        # [69, 1.1],
        [78, 1.1],
        # No tree crops hit clamp of 1.2 with an NDVI of 0.9
        # Peaches only hit with NDVI of 0.95
        [67, 1.2],
    ]
)
def test_Model_kc_crop_type_kc_clamping(crop_type_source, expected):
    m = default_model_obj(crop_type_source=crop_type_source, crop_type_kc_flag=True)
    output = utils.constant_image_value(m.kc(ndvi=ee.Image.constant(0.95)))
    assert output['kc'] == expected


@pytest.mark.parametrize(
    'crop_type, ndvi, water_kc_flag, expected',
    [
        [0, -0.2, False, 0.0],
        [0, -0.2, True, 1.05],
        [1, -0.2, True, 0.15],  # Only set water Kc on crop_class 0
    ]
)
def test_Model_kc_water_kc_flag(crop_type, ndvi, water_kc_flag, expected):
    m = default_model_obj(
        crop_type_source=crop_type, crop_type_kc_flag=False, water_kc_flag=water_kc_flag
    )
    output = utils.constant_image_value(m.kc(ndvi=ee.Image.constant(ndvi)))
    assert output['kc'] == expected


def ndvi_to_kc_point(ndvi, doy, crop_type):
    crop_profile = data.cdl[crop_type]

    fc = min(max((1.26 * ndvi) - 0.18, 0), 1)
    # print(crop_profile['crop_class'])
    if crop_profile['crop_class'] == 1:
        h = crop_profile['h_max'] * min((fc / 0.7), 1)
        fr = 1.0
    elif crop_profile['crop_class'] == 3 or crop_profile['crop_class'] == 2:
        # Set fr based on doy
        if doy < crop_profile['ls_start']:
            fr = crop_profile['fr_mid']
        elif crop_profile['ls_start'] <= doy and doy <= crop_profile['ls_stop']:
            fr = crop_profile["fr_mid"] - (
                (doy - crop_profile["ls_start"])
                / (crop_profile["ls_stop"] - crop_profile["ls_start"])
                * (crop_profile["fr_mid"] - crop_profile["fr_end"])
            )
        elif doy > crop_profile['ls_stop']:
            fr = crop_profile['fr_end']

        # Set h based on crop class
        if crop_profile['crop_class'] == 3:
            if fc > 0.5:
                h = crop_profile['h_max']
            else:
                h = crop_profile['h_max'] - 1
        elif crop_profile['crop_class'] == 2:
            h = crop_profile['h_max']
    else:
        return -1

    kd = min(1, crop_profile['m_l'] * fc, fc ** (1 / (1 + h)))
    kcb_full = fr * min(1 + (0.1 * crop_profile['h_max']), 1.2)
    kc_min = 0.15
    kcb = kc_min + kd * (kcb_full - kc_min)

    # Crop class ceilings
    if crop_profile['crop_class'] == 2:
        kcb = min(kcb, 1.1)
    elif crop_profile['crop_class'] == 3:
        kcb = min(kcb, 1.2)

    return kcb


@pytest.mark.parametrize(
    'ndvi, doy, crop_type_num',
    [
        # 1.26 * 0.8 - 0.18 = 0.828
        # ((0.828 ** 2) * -0.4771) + (1.4047 * 0.828) + 0.15 = 0.9859994736
        [0.8, 174, 1],
        # [1.0, 200, 3],
        # [0.5, 200, 3],
        # [0.1, 200, 3],
        [1.0, 200, 1],
        [0.5, 200, 1],
        [0.1, 200, 1],
        [1.0, 200, 2],
        [0.5, 250, 2],
        [0.1, 300, 2],
        [1.0, 200, 69],
        [0.5, 200, 69],
        [0.1, 200, 69],
        [1.0, 250, 69],
        [0.5, 250, 69],
        [0.1, 250, 69],
        [1.0, 300, 69],
        [0.5, 300, 69],
        [0.1, 300, 69],
        [1.0, 200, 75],
        [0.5, 200, 75],
        [0.1, 200, 75],
        [1.0, 300, 75],
        [0.5, 300, 75],
        [0.1, 300, 75],
        [1.0, 301, 75],
        [0.5, 301, 75],
        [0.1, 301, 75],
    ]
)
def test_Image_kc_constant_value(ndvi, doy, crop_type_num, tol=0.0001):
    ndvi_img = ee.Image.constant(ndvi)
    m = default_model_obj(crop_type_source=crop_type_num, doy=doy, crop_type_kc_flag=True)
    output = utils.constant_image_value(m.kc(ndvi_img))
    expected = ndvi_to_kc_point(ndvi, doy, crop_type_num)
    assert abs(output['kc'] - expected) <= tol


# Check the Kcb calculation for alfalfa
@pytest.mark.parametrize(
    'crop_type, fc, expected',
    [
        # [36, -0.1, 0.15],   # Check clamping
        [36, 0.0, 0.15],  # NDVI == 0.1429
        [36, 0.1, 0.2634],  # NDVI == 0.2222
        [36, 0.2, 0.3906],  # NDVI == 0.3016
        [36, 0.3, 0.5144],  # NDVI == 0.3809
        [36, 0.4, 0.6281],  # NDVI == 0.4603
        [36, 0.5, 0.7296],  # NDVI == 0.5397
        [36, 0.6, 0.8185],  # NDVI == 0.6190
        [36, 0.7, 0.8959],  # NDVI == 0.6984
        [36, 0.8, 0.9568],  # NDVI == 0.7778
        [36, 0.9, 1.0147],  # NDVI == 0.8571
        [36, 1.0, 1.07],
        [36, 1.1, 1.07],  # Check clamping
    ]
)
def test_Model_crop_type_kcb(crop_type, fc, expected, tol=0.0001):
    m = default_model_obj(crop_type_source=crop_type, crop_type_kc_flag=True)
    output = utils.constant_image_value(m._kcb(m._kd_row_crop(fc=ee.Image.constant(fc))))
    assert abs(output['kcb'] - expected) <= tol


@pytest.mark.parametrize(
    'crop_type, fc, expected',
    [
        # [36, -0.1, 0.0],   # Check clamping
        [36, 0.0, 0.0],  # NDVI == 0.1429
        [36, 0.1, 0.1233],  # NDVI == 0.2222
        [36, 0.2, 0.2615],  # NDVI == 0.3016
        [36, 0.3, 0.3961],  # NDVI == 0.3809
        [36, 0.4, 0.5197],  # NDVI == 0.4603
        [36, 0.5, 0.6300],  # NDVI == 0.5397
        [36, 0.6, 0.7267],  # NDVI == 0.6190
        [36, 0.7, 0.8107],  # NDVI == 0.6984
        [36, 0.8, 0.8770],  # NDVI == 0.7778
        [36, 0.9, 0.9399],  # NDVI == 0.8571
        [36, 1.0, 1.0],
        [36, 1.1, 1.0],  # Check clamping
    ]
)
def test_Model_crop_type_kd(crop_type, fc, expected, tol=0.0001):
    m = default_model_obj(crop_type_source=crop_type, crop_type_kc_flag=True)
    output = utils.constant_image_value(m._kd_row_crop(fc=ee.Image.constant(fc)))
    assert abs(output['kd'] - expected) <= tol



import copy
import datetime
from importlib import metadata
# import pprint

from dateutil.relativedelta import relativedelta
import ee
import openet.core.interpolate
# TODO: import utils from openet.core
# import openet.core.utils as utils

from . import utils
from .image import Image


def lazy_property(fn):
    """Decorator that makes a property lazy-evaluated

    https://stevenloria.com/lazy-properties/
    """
    attr_name = '_lazy_' + fn.__name__

    @property
    def _lazy_property(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, fn(self))
        return getattr(self, attr_name)
    return _lazy_property


class Collection():
    """"""

    def __init__(
        self,
        collections,
        start_date,
        end_date,
        geometry,
        variables=None,
        cloud_cover_max=70,
        et_reference_source=None,
        et_reference_band=None,
        et_reference_factor=None,
        et_reference_resample=None,
        filter_args=None,
        model_args=None,
        # model_args={'et_reference_source': 'IDAHO_EPSCOR/GRIDMET',
        #             'et_reference_band': 'eto',
        #             'et_reference_factor': 0.85,
        #             'et_reference_resample': 'nearest},
        # **kwargs
    ):
        """Earth Engine based SIMS ETcb Image Collection object

        Parameters
        ----------
        collections : list, str
            GEE satellite image collection IDs.
        start_date : str
            ISO format inclusive start date (i.e. YYYY-MM-DD).
        end_date : str
            ISO format exclusive end date (i.e. YYYY-MM-DD).
            This date needs to be exclusive since it will be passed directly
            to the .filterDate() calls.
        geometry : ee.Geometry
            The geometry object will be used to filter the input collections
            using the ee.ImageCollection.filterBounds() method.
        variables : list, optional
            Output variables can also be specified in the method calls.
        cloud_cover_max : float
            Maximum cloud cover percentage (the default is 70%).
                - Landsat SR: CLOUD_COVER_LAND
                - Sentinel2: CLOUDY_PIXEL_PERCENTAGE
        et_reference_source : str, float, optional
            Reference ET source (the default is None).  ETr Parameters must
            be be set here or in model args to interpolate ET, ETf, or ETr.
        et_reference_band : str, optional
            Reference ET band name (the default is None).  ETr Parameters must
            be be set here or in model args to interpolate ET, ETf, or ETr.
        et_reference_factor : float, None, optional
            Reference ET scaling factor.  The default is None which is
            equivalent to 1.0 (or no scaling).
        et_reference_resample : {'nearest', 'bilinear', 'bicubic'}, None, optional
            Reference ET resampling.  The default is None which is equivalent
            to nearest neighbor resampling.
        filter_args : dict
            Image collection filter keyword arguments (the default is None).
            Organize filter arguments as a nested dictionary with the primary
            key being the collection ID.
        model_args : dict
            Model Image initialization keyword arguments (the default is None).
            Dictionary will be passed through to model Image init.

        """
        self.collections = collections
        self.variables = variables
        self.start_date = start_date
        self.end_date = end_date
        self.geometry = geometry
        self.cloud_cover_max = cloud_cover_max

        # CGM - Should we check that model_args and filter_args are dict?
        if model_args is not None:
            self.model_args = model_args
        else:
            self.model_args = {}
        if filter_args is not None:
            self.filter_args = filter_args
        else:
            self.filter_args = {}

        # Reference ET parameters
        self.et_reference_source = et_reference_source
        self.et_reference_band = et_reference_band
        self.et_reference_factor = et_reference_factor
        self.et_reference_resample = et_reference_resample

        # Check reference ET parameters
        if et_reference_factor and not utils.is_number(et_reference_factor):
            raise ValueError('et_reference_factor must be a number')
        if et_reference_factor and (self.et_reference_factor < 0):
            raise ValueError('et_reference_factor must be greater than zero')
        et_reference_resample_methods = ['nearest', 'bilinear', 'bicubic']
        if (et_reference_resample and
                (et_reference_resample.lower() not in et_reference_resample_methods)):
            raise ValueError('unsupported et_reference_resample method')

        # Set/update the ETr parameters in model_args if they were set in init()
        if self.et_reference_source:
            self.model_args['et_reference_source'] = self.et_reference_source
        if self.et_reference_band:
            self.model_args['et_reference_band'] = self.et_reference_band
        if self.et_reference_factor:
            self.model_args['et_reference_factor'] = self.et_reference_factor
        if self.et_reference_resample:
            self.model_args['et_reference_resample'] = self.et_reference_resample

        # Model specific variables that can be interpolated to a daily timestep
        # CGM - Should this be specified in the interpolation method instead?
        self._interp_vars = ['ndvi', 'et_fraction']

        self._landsat_c2_sr_collections = [
            'LANDSAT/LT04/C02/T1_L2',
            'LANDSAT/LT05/C02/T1_L2',
            'LANDSAT/LE07/C02/T1_L2',
            'LANDSAT/LC08/C02/T1_L2',
            'LANDSAT/LC09/C02/T1_L2',
        ]

        # If collections is a string, place in a list
        if type(self.collections) is str:
            self.collections = [self.collections]

        # Check that collection IDs are supported
        for coll_id in self.collections:
            if coll_id not in self._landsat_c2_sr_collections:
                raise ValueError(f'unsupported collection: {coll_id}')

        # CGM - This test is not needed since only Landsat SR collections are supported
        # # Check that collections don't have "duplicates"
        # #   (i.e TOA and SR or TOA and TOA_RT for same Landsat)
        # def duplicates(x):
        #     return len(x) != len(set(x))
        # if duplicates([c.split('/')[1] for c in self.collections]):
        #     raise ValueError('duplicate landsat types in collection list')

        # Check start/end date
        if not utils.valid_date(self.start_date):
            raise ValueError('start_date is not a valid')
        elif not utils.valid_date(self.end_date):
            raise ValueError('end_date is not valid')
        elif not self.start_date < self.end_date:
            raise ValueError('end_date must be after start_date')

        # Check cloud_cover_max
        if ((type(self.cloud_cover_max) is not int) and
                (type(self.cloud_cover_max) is not float) and
                not utils.is_number(self.cloud_cover_max)):
            raise TypeError('cloud_cover_max must be a number')
        if ((type(self.cloud_cover_max) is str) and
                utils.is_number(self.cloud_cover_max)):
            self.cloud_cover_max = float(self.cloud_cover_max)
        if (self.cloud_cover_max < 0) or (self.cloud_cover_max > 100):
            raise ValueError('cloud_cover_max must be in the range 0 to 100')

        # Check geometry?
        # if not isinstance(self.geometry, computedobject.ComputedObject):
        #     raise ValueError()

        # Filter collection list based on start/end dates
        if self.end_date <= '1982-01-01':
            self.collections = [c for c in self.collections if 'LT04' not in c]
        if self.start_date >= '1994-01-01':
            self.collections = [c for c in self.collections if 'LT04' not in c]
        if self.end_date <= '1984-01-01':
            self.collections = [c for c in self.collections if 'LT05' not in c]
        if self.start_date >= '2012-01-01':
            self.collections = [c for c in self.collections if 'LT05' not in c]
        if self.end_date <= '1999-01-01':
            self.collections = [c for c in self.collections if 'LE07' not in c]
        if self.start_date >= '2022-01-01':
            self.collections = [c for c in self.collections if 'LE07' not in c]
        if self.end_date <= '2013-01-01':
            self.collections = [c for c in self.collections if 'LC08' not in c]
        if self.end_date <= '2022-01-01':
            self.collections = [c for c in self.collections if 'LC09' not in c]
        # if self.end_date <= '2015-01-01':
        #     self.collections = [c for c in self.collections if 'COPERNICUS' not in c]

    def _build(self, variables=None, start_date=None, end_date=None):
        """Build a merged model variable image collection

        Parameters
        ----------
        variables : list
            Set a variable list that is different than the class variable list.
        start_date : str, optional
            Set a start_date that is different than the class start_date.
            This is needed when defining the scene collection to have extra
            images for interpolation.
        end_date : str, optional
            Set an exclusive end_date that is different than the class end_date.

        Returns
        -------
        ee.ImageCollection

        Raises
        ------
        ValueError if collection IDs are invalid.
        ValueError if variables is not set here and in class init.

        """
        # Override the class parameters if necessary
        # Distinguish between variables defaulting to None, and variables being
        #   set to an empty list, in which case the merged landsat collection
        #   should be returned.
        if variables is None:
            if self.variables:
                variables = self.variables
            else:
                raise ValueError('variables parameter must be set')
        elif not variables:
            pass
        if not start_date:
            start_date = self.start_date
        if not end_date:
            end_date = self.end_date

        # Build the variable image collection
        variable_coll = ee.ImageCollection([])
        for coll_id in self.collections:
            # TODO: Move to separate methods/functions for each type
            if coll_id in self._landsat_c2_sr_collections:
                input_coll = (
                    ee.ImageCollection(coll_id)
                    .filterDate(start_date, end_date)
                    .filterBounds(self.geometry)
                    .filterMetadata('CLOUD_COVER_LAND', 'less_than', self.cloud_cover_max)
                    .filterMetadata('CLOUD_COVER_LAND', 'greater_than', -0.5)
                )

                # TODO: Move this to a separate function (maybe in utils.py?)
                #   since it is identical for all the supported collections
                if ((self.filter_args is None) or
                        (not isinstance(self.filter_args, dict)) or
                        (coll_id not in self.filter_args.keys())):
                    pass
                elif isinstance(self.filter_args[coll_id], ee.ComputedObject):
                    input_coll = input_coll.filter(self.filter_args[coll_id])
                elif isinstance(self.filter_args[coll_id], list):
                    # TODO: This generic dictionary based filtering should
                    #   probably be removed since only the "equals" filter
                    #   has been implemented and the functionality is better
                    #   handled with the other approach.
                    for f in copy.deepcopy(self.filter_args[coll_id]):
                        try:
                            filter_type = f.pop('type')
                        except KeyError:
                            continue
                        if filter_type.lower() == 'equals':
                            input_coll = input_coll.filter(ee.Filter.equals(**f))
                else:
                    raise ValueError('Unsupported filter_arg parameter')

                # Time filters are to remove bad (L5) and pre-op (L8) images
                if 'LT05' in coll_id:
                    input_coll = input_coll.filter(ee.Filter.lt(
                        'system:time_start', ee.Date('2011-12-31').millis()
                    ))
                elif 'LE07' in coll_id:
                    input_coll = input_coll.filter(ee.Filter.lt(
                        'system:time_start', ee.Date('2022-01-01').millis()
                    ))
                elif 'LC08' in coll_id:
                    input_coll = input_coll.filter(ee.Filter.gt(
                        'system:time_start', ee.Date('2013-04-01').millis()
                    ))
                elif 'LC09' in coll_id:
                    input_coll = input_coll.filter(ee.Filter.gt(
                        'system:time_start', ee.Date('2022-01-01').millis()
                    ))

                def compute_vars(image):
                    model_obj = Image.from_landsat_c2_sr(
                        sr_image=ee.Image(image), **self.model_args
                    )
                    return model_obj.calculate(variables)

                # Skip going into image class if variables is not set so raw
                #   landsat collection can be returned for getting image_id_list
                if variables:
                    input_coll = ee.ImageCollection(input_coll.map(compute_vars))

                variable_coll = variable_coll.merge(input_coll)

            else:
                raise ValueError(f'unsupported collection: {coll_id}')

        return variable_coll

    def overpass(self, variables=None):
        """Return a collection of computed values for the overpass images

        Parameters
        ----------
        variables : list, optional
            List of variables that will be returned in the Image Collection.
            If variables is not set here it must be specified in the class
            instantiation call.

        Returns
        -------
        ee.ImageCollection

        Raises
        ------
        ValueError

        """
        # Does it make sense to use the class variable list if not set?
        if not variables:
            if self.variables:
                variables = self.variables
            else:
                raise ValueError('variables parameter must be set')

        return self._build(variables=variables)

    def interpolate(
            self,
            variables=None,
            t_interval='custom',
            interp_method='linear',
            interp_days=32,
            use_joins=True,
            mask_partial_aggregations=True,
            **kwargs,
    ):
        """

        Parameters
        ----------
        variables : list, optional
            List of variables that will be returned in the Image Collection.
            If variables is not set here it must be specified in the class
            instantiation call.
        t_interval : {'daily', 'monthly', 'custom'}, optional
            Time interval over which to interpolate and aggregate values
            The default 'custom' interval will aggregate all days within the
            start/end dates and return an image collection with a single image.
        interp_method : {'linear}, optional
            Interpolation method (the default is 'linear').
        interp_days : int, str, optional
            Number of extra days before the start date and after the end date
            to include in the interpolation calculation. (the default is 32).
        use_joins : bool, optional
            If True, use joins to link the target and source collections.
            If False, the source collection will be filtered for each target image.
            This parameter is passed through to interpolate.daily().
        mask_partial_aggregations : bool, optional
            If True, pixels with an aggregation count less than the number of
            days in the aggregation time period will be masked.  The default is True.
        kwargs : dict, optional

        Returns
        -------
        ee.ImageCollection

        Raises
        ------
        ValueError for unsupported input parameters
        ValueError for negative interp_days values
        TypeError for non-integer interp_days

        Notes
        -----
        Not all variables can be interpolated to new time steps.
        Variables like reference ET are simply summed whereas ET fraction is
            computed from the interpolated/aggregated values.

        """
        # Check that the input parameters are valid
        if t_interval.lower() not in ['daily', 'monthly', 'custom']:
            raise ValueError(f'unsupported t_interval: {t_interval}')
        elif interp_method.lower() not in ['linear']:
            raise ValueError(f'unsupported interp_method: {interp_method}')

        if (type(interp_days) is str) and utils.is_number(interp_days):
            interp_days = int(interp_days)
        elif not type(interp_days) is int:
            raise TypeError('interp_days must be an integer')
        elif interp_days <= 0:
            raise ValueError('interp_days must be a positive integer')

        # Does it make sense to use the class variable list if not set?
        if not variables:
            if self.variables:
                variables = self.variables
            else:
                raise ValueError('variables parameter must be set')

        # Adjust start/end dates based on t_interval
        # Increase the date range to fully include the time interval
        start_dt = datetime.datetime.strptime(self.start_date, '%Y-%m-%d')
        end_dt = datetime.datetime.strptime(self.end_date, '%Y-%m-%d')
        if t_interval.lower() == 'monthly':
            start_dt = datetime.datetime(start_dt.year, start_dt.month, 1)
            end_dt -= relativedelta(days=+1)
            end_dt = datetime.datetime(end_dt.year, end_dt.month, 1)
            end_dt += relativedelta(months=+1)
        # elif t_interval.lower() == 'annual':
        #     start_dt = datetime.datetime(start_dt.year, 1, 1)
        #     # Covert end date to inclusive, flatten to beginning of year,
        #     # then add a year which will make it exclusive
        #     end_dt -= relativedelta(days=+1)
        #     end_dt = datetime.datetime(end_dt.year, 1, 1)
        #     end_dt += relativedelta(years=+1)
        start_date = start_dt.strftime('%Y-%m-%d')
        end_date = end_dt.strftime('%Y-%m-%d')

        # The start/end date for the interpolation include more days
        # (+/- interp_days) than are included in the ETr collection
        interp_start_dt = start_dt - datetime.timedelta(days=interp_days)
        interp_end_dt = end_dt + datetime.timedelta(days=interp_days)
        interp_start_date = interp_start_dt.date().isoformat()
        interp_end_date = interp_end_dt.date().isoformat()

        # Update model_args if et_reference parameters were passed to interpolate
        # Intentionally using model_args (instead of self.et_reference_source, etc.) in
        #   this function since model_args is passed to Image class in _build()
        # if ('et' in variables) or ('et_reference' in variables):
        if (('et_reference_source' in kwargs.keys()) and
                (kwargs['et_reference_source'] is not None)):
            self.model_args['et_reference_source'] = kwargs['et_reference_source']
        if (('et_reference_band' in kwargs.keys()) and
                (kwargs['et_reference_band'] is not None)):
            self.model_args['et_reference_band'] = kwargs['et_reference_band']
        if (('et_reference_factor' in kwargs.keys()) and
                (kwargs['et_reference_factor'] is not None)):
            self.model_args['et_reference_factor'] = kwargs['et_reference_factor']
        if (('et_reference_resample' in kwargs.keys()) and
                (kwargs['et_reference_resample'] is not None)):
            self.model_args['et_reference_resample'] = kwargs['et_reference_resample'].lower()

        # Check that all et_reference parameters were set
        for et_reference_param in ['et_reference_source', 'et_reference_band']:
            if et_reference_param not in self.model_args.keys():
                raise ValueError(f'{et_reference_param} was not set')
            elif not self.model_args[et_reference_param]:
                raise ValueError(f'{et_reference_param} was not set')

        if type(self.model_args['et_reference_source']) is str:
            # Assume a string source is a single image collection ID
            #   not a list of collection IDs or ee.ImageCollection
            daily_et_ref_coll = (
                ee.ImageCollection(self.model_args['et_reference_source'])
                .filterDate(start_date, end_date)
                .select([self.model_args['et_reference_band']], ['et_reference'])
            )
        # elif isinstance(self.model_args['et_reference_source'], computedobject.ComputedObject):
        #     # Interpret computed objects as image collections
        #     daily_et_ref_coll = (
        #         self.model_args['et_reference_source']
        #         .filterDate(self.start_date, self.end_date)
        #         .select([self.model_args['et_reference_band']])
        #     )
        else:
            raise ValueError(f'unsupported et_reference_source: '
                             f'{self.model_args["et_reference_source"]}')

        # Scale reference ET images (if necessary)
        # Support for applying an adjustment factor to the reference ET
        #   may be removed at some point
        if (('et_reference_factor' in self.model_args.keys()) and
                self.model_args['et_reference_factor'] and
                (self.model_args['et_reference_factor'] != 1) and
                (self.model_args['et_reference_factor'] > 0)):
            def et_reference_adjust(input_img):
                return (
                    input_img.multiply(self.model_args['et_reference_factor'])
                    .copyProperties(input_img)
                    .set({'system:time_start': input_img.get('system:time_start')})
                )
            daily_et_ref_coll = daily_et_ref_coll.map(et_reference_adjust)

        # Initialize variable list to only variables that can be interpolated
        interp_vars = list(set(self._interp_vars) & set(variables))

        # To return ET, the ET fraction must be interpolated
        if ('et' in variables) and ('et_fraction' not in interp_vars):
            interp_vars = interp_vars + ['et_fraction']

        # With the current interpolate.daily() function,
        #   something has to be interpolated in order to return et_reference
        if ('et_reference' in variables) and ('et_fraction' not in interp_vars):
            interp_vars = interp_vars + ['et_fraction']

        # The time band is always needed for interpolation
        interp_vars = interp_vars + ['time']

        # Count will be determined using the aggregate_coll image masks
        if 'count' in variables:
            interp_vars = interp_vars + ['mask']
            # interp_vars.remove('count')

        # Build initial scene image collection
        scene_coll = self._build(
            variables=interp_vars,
            start_date=interp_start_date,
            end_date=interp_end_date,
        )

        # For count, compute the composite/mosaic image for the mask band only
        if 'count' in variables:
            aggregate_coll = openet.core.interpolate.aggregate_to_daily(
                image_coll=scene_coll.select(['mask']),
                start_date=start_date,
                end_date=end_date,
            )

            # The following is needed because the aggregate collection can be
            #   empty if there are no scenes in the target date range but there
            #   are scenes in the interpolation date range.
            # Without this the count image will not be built but the other
            #   bands will be which causes a non-homogenous image collection.
            aggregate_coll = aggregate_coll.merge(
                ee.Image.constant(0).rename(['mask'])
                .set({'system:time_start': ee.Date(start_date).millis()})
            )

        # Including count/mask causes problems in interpolate.daily() function.
        # Issues with mask being an int but the values need to be double.
        # Casting the mask band to a double would fix this problem also.
        if 'mask' in interp_vars:
            interp_vars.remove('mask')

        # Interpolate to a daily time step
        # NOTE: the daily function is not computing ET (ETf x ETr)
        #   but is returning the target (ETr) band
        daily_coll = openet.core.interpolate.daily(
            target_coll=daily_et_ref_coll,
            source_coll=scene_coll.select(interp_vars),
            interp_method=interp_method,
            interp_days=interp_days,
            use_joins=use_joins,
            compute_product=False,
            # resample_method=et_reference_resample,
        )

        # Compute ET from ET fraction and reference ET (if necessary)
        # CGM - The conditional is needed if only interpolating NDVI
        if ('et' in variables) or ('et_fraction' in variables):
            def compute_et(img):
                """This function assumes et_reference and et_fraction are present"""
                # Apply any resampling to the reference ET image before computing ET
                et_ref_img = img.select(['et_reference'])
                if (self.model_args['et_reference_resample'] and
                        (self.model_args['et_reference_resample'] in ['bilinear', 'bicubic'])):
                    et_ref_img = et_ref_img.resample(self.model_args['et_reference_resample'])

                et_img = img.select(['et_fraction']).multiply(et_ref_img)

                return img.addBands(et_img.double().rename('et'))

            daily_coll = daily_coll.map(compute_et)

        interp_properties = {
            'cloud_cover_max': self.cloud_cover_max,
            'collections': ', '.join(self.collections),
            'interp_days': interp_days,
            'interp_method': interp_method,
            'model_name': metadata.metadata('openet-sims')['Name'],
            'model_version': metadata.metadata('openet-sims')['Version'],
            # 'model_name': openet.sims.MODEL_NAME,
            # 'model_version': openet.sims.__version__,
        }
        interp_properties.update(self.model_args)

        # CGM - This function is being declared here to avoid passing in all the common parameters
        #   such as: daily_coll, daily_et_ref_coll, interp_properties, variables, etc.
        # Long term it should probably be declared outside of this function
        #   or read from openet-core
        def aggregate_image(agg_start_date, agg_end_date, date_format):
            """Aggregate the daily images within the target date range

            Parameters
            ----------
            agg_start_date: str
                Start date (inclusive).
            agg_end_date : str
                End date (exclusive).
            date_format : str
                Date format for system:index (uses EE JODA format).

            Returns
            -------
            ee.Image

            """
            et_img = None
            eto_img = None

            if ('et' in variables) or ('et_fraction' in variables):
                et_img = daily_coll.filterDate(agg_start_date, agg_end_date).select(['et']).sum()

            if ('et_reference' in variables) or ('et_fraction' in variables):
                eto_img = (
                    daily_coll.filterDate(agg_start_date, agg_end_date)
                    .select(['et_reference']).sum()
                )
                if (self.model_args['et_reference_resample'] and
                        (self.model_args['et_reference_resample'] in ['bilinear', 'bicubic'])):
                    eto_img = (
                        eto_img.setDefaultProjection(daily_et_ref_coll.first().projection())
                        .resample(self.model_args['et_reference_resample'])
                    )

            # Count the number of interpolated/aggregated values
            # Mask pixels that do not have a full aggregation count for the start/end
            if ('et' in variables) or ('et_fraction' in variables):
                aggregation_band = 'et'
            elif 'ndvi' in interp_vars:
                aggregation_band = 'ndvi'
            else:
                raise ValueError('no supported aggregation band')
            aggregation_count_img = (
                daily_coll.filterDate(agg_start_date, agg_end_date)
                .select([aggregation_band]).reduce(ee.Reducer.count())
            )

            image_list = []
            if 'et' in variables:
                image_list.append(et_img.float())
            if 'et_reference' in variables:
                image_list.append(eto_img.float())
            if 'et_fraction' in variables:
                # Average et fraction over the aggregation period
                image_list.append(et_img.divide(eto_img).rename(['et_fraction']).float())
            if 'ndvi' in variables:
                # Average NDVI over the aggregation period
                ndvi_img = (
                    daily_coll.filterDate(agg_start_date, agg_end_date)
                    .select(['ndvi']).mean().float()

                )
                image_list.append(ndvi_img)
            if ('scene_count' in variables) or ('count' in variables):
                scene_count_img = (
                    aggregate_coll.filterDate(agg_start_date, agg_end_date)
                    .select(['mask']).reduce(ee.Reducer.sum()).rename('count').uint8()
                )
                image_list.append(scene_count_img)
            if 'daily_count' in variables:
                image_list.append(aggregation_count_img.rename('daily_count').uint8())

            output_img = ee.Image(image_list)

            if mask_partial_aggregations:
                aggregation_days = ee.Date(agg_end_date).difference(ee.Date(agg_start_date), 'day')
                aggregation_count_mask = aggregation_count_img.gte(aggregation_days.subtract(1))
                # aggregation_count_mask = agg_count_img.gte(aggregation_days)
                output_img = output_img.updateMask(aggregation_count_mask)

            return (
                output_img
                .set({
                    'system:index': ee.Date(agg_start_date).format(date_format),
                    'system:time_start': ee.Date(agg_start_date).millis(),
                })
                .set(interp_properties)
            )

        # Combine input, interpolated, and derived values
        if t_interval.lower() == 'custom':
            # Return an ImageCollection to be consistent with the other t_interval options
            return ee.ImageCollection(aggregate_image(
                agg_start_date=start_date,
                agg_end_date=end_date,
                date_format='YYYYMMdd',
            ))
        elif t_interval.lower() == 'daily':
            def aggregate_daily(daily_img):
                # CGM - Double check that this time_start is a 0 UTC time.
                # It should be since it is coming from the interpolate source
                #   collection, but what if source is GRIDMET (+6 UTC)?
                agg_start_date = ee.Date(daily_img.get('system:time_start'))
                # CGM - This calls .sum() on collections with only one image
                return aggregate_image(
                    agg_start_date=agg_start_date,
                    agg_end_date=ee.Date(agg_start_date).advance(1, 'day'),
                    date_format='YYYYMMdd',
                )
            return ee.ImageCollection(daily_coll.map(aggregate_daily))
        elif t_interval.lower() == 'monthly':
            def month_gen(iter_start_dt, iter_end_dt):
                iter_dt = iter_start_dt
                # Conditional is "less than" because end date is exclusive
                while iter_dt < iter_end_dt:
                    yield iter_dt.strftime('%Y-%m-%d')
                    iter_dt += relativedelta(months=+1)
            def aggregate_monthly(agg_start_date):
                return aggregate_image(
                    agg_start_date=agg_start_date,
                    agg_end_date=ee.Date(agg_start_date).advance(1, 'month'),
                    date_format='YYYYMM',
                )
            return ee.ImageCollection(
                ee.List(list(month_gen(start_dt, end_dt))).map(aggregate_monthly)
            )
        else:
            raise ValueError(f'unsupported t_interval: {t_interval}')

    def get_image_ids(self):
        """Return image IDs of the input images

        Note, this does not return the extra images used for interpolation

        Returns
        -------
        list

        """
        # CGM - Setting variables to None bypasses the Image class, so image_id
        #   is not set and merge indices must be removed from the system:index
        return list(utils.getinfo(self._build(variables=[]).aggregate_array('system:id')))
        # return list(utils.getinfo(self._build(variables=['ndvi']).aggregate_array('image_id')))



import datetime
# import logging
# import pprint

import ee
import pytest

import openet.sims as sims
import openet.sims.utils as utils
# TODO: import utils from openet.core
# import openet.core.utils as utils


COLL_ID = 'LANDSAT/LC08/C02/T1_L2/'
SCENE_ID = 'LC08_044033_20170716'
SCENE_DT = datetime.datetime.strptime(SCENE_ID[-8:], '%Y%m%d')
SCENE_DATE = SCENE_DT.strftime('%Y-%m-%d')
SCENE_DOY = int(SCENE_DT.strftime('%j'))
SCENE_TIME = utils.millis(SCENE_DT)
# SCENE_TIME = utils.getinfo(ee.Date(SCENE_DATE).millis())


# Should these be test fixtures instead?
# I'm not sure how to make them fixtures and allow input parameters
def input_image(red=0.1, nir=0.9):
    """Construct a fake input image with renamed bands"""
    return ee.Image.constant([red, nir]).rename(['red', 'nir'])\
        .set({'system:time_start': ee.Date(SCENE_DATE).millis()})


def default_image(ndvi=0.8):
    return ee.Image.constant([ndvi]).rename(['ndvi'])\
        .set({
            'system:index': SCENE_ID,
            'system:time_start': ee.Date(SCENE_DATE).millis(),
            'system:id': COLL_ID + SCENE_ID,
        })


# Setting et_reference_source and et_reference_band on the default image to
# simplify testing but these do not have defaults in the Image class init
def default_image_args(
        ndvi=0.8,
        et_reference_source='IDAHO_EPSCOR/GRIDMET',
        et_reference_band='etr',
        et_reference_factor=0.85,
        et_reference_resample='nearest',
        crop_type_source='USDA/NASS/CDL',
        crop_type_remap='CDL',
        crop_type_kc_flag=False,
        crop_type_annual_skip_flag=False,
        mask_non_ag_flag=False,
        water_kc_flag=True,
        # reflectance_type='SR',
        ):
    return {
        'image': default_image(ndvi=ndvi),
        'et_reference_source': et_reference_source,
        'et_reference_band': et_reference_band,
        'et_reference_factor': et_reference_factor,
        'et_reference_resample': et_reference_resample,
        'crop_type_source': crop_type_source,
        'crop_type_remap': crop_type_remap,
        'crop_type_kc_flag': crop_type_kc_flag,
        'crop_type_annual_skip_flag': crop_type_annual_skip_flag,
        'mask_non_ag_flag': mask_non_ag_flag,
        'water_kc_flag': water_kc_flag,
        # 'reflectance_type': reflectance_type,
    }


def default_image_obj(
        ndvi=0.8,
        et_reference_source='IDAHO_EPSCOR/GRIDMET',
        et_reference_band='etr',
        et_reference_factor=0.85,
        et_reference_resample='nearest',
        crop_type_source='USDA/NASS/CDL',
        crop_type_remap='CDL',
        crop_type_kc_flag=False,
        crop_type_annual_skip_flag=False,
        mask_non_ag_flag=False,
        water_kc_flag=True,
        # reflectance_type='SR',
        ):
    return sims.Image(**default_image_args(
        ndvi=ndvi,
        et_reference_source=et_reference_source,
        et_reference_band=et_reference_band,
        et_reference_factor=et_reference_factor,
        et_reference_resample=et_reference_resample,
        crop_type_source=crop_type_source,
        crop_type_remap=crop_type_remap,
        crop_type_kc_flag=crop_type_kc_flag,
        crop_type_annual_skip_flag=crop_type_annual_skip_flag,
        mask_non_ag_flag=mask_non_ag_flag,
        water_kc_flag=water_kc_flag,
        # reflectance_type=reflectance_type,
    ))


def test_Image_init_default_parameters():
    m = sims.Image(image=default_image())
    assert m.et_reference_source is None
    assert m.et_reference_band is None
    assert m.et_reference_factor is None
    assert m.et_reference_resample is None
    # assert m.crop_type_source == 'USDA/NASS/CDL'
    # assert m.crop_type_remap == 'CDL'
    # assert m.crop_type_kc_flag == False
    # assert m.crop_type_annual_skip_flag == False
    # assert m.mask_non_ag_flag == False
    assert m.reflectance_type == 'SR'


def test_Image_init_calculated_properties():
    m = default_image_obj()
    assert utils.getinfo(m._time_start) == SCENE_TIME
    assert utils.getinfo(m._index) == SCENE_ID


def test_Image_init_date_properties():
    m = default_image_obj()
    assert utils.getinfo(m._date)['value'] == SCENE_TIME
    assert utils.getinfo(m._year) == int(SCENE_DATE.split('-')[0])
    assert utils.getinfo(m._start_date)['value'] == SCENE_TIME
    assert utils.getinfo(m._end_date)['value'] == utils.millis(
        SCENE_DT + datetime.timedelta(days=1))


@pytest.mark.parametrize(
    'red, nir, expected',
    [
        [0.2, 9.0 / 55, -0.1],
        [0.2, 0.2,  0.0],
        [0.1, 11.0 / 90,  0.1],
        [0.2, 0.3, 0.2],
        [0.1, 13.0 / 70, 0.3],
        [0.3, 0.7, 0.4],
        [0.2, 0.6, 0.5],
        [0.2, 0.8, 0.6],
        [0.1, 17.0 / 30, 0.7],
        [0.1, 0.9, 0.8],
    ]
)
def test_Image_static_ndvi_calculation(red, nir, expected, tol=0.000001):
    output = utils.constant_image_value(sims.Image._ndvi(input_image(red=red, nir=nir)))
    assert abs(output['ndvi'] - expected) <= tol


def test_Image_static_ndvi_band_name():
    output = utils.getinfo(sims.Image._ndvi(input_image()))
    assert output['bands'][0]['id'] == 'ndvi'


def test_Image_ndvi_properties():
    """Test if properties are set on the ndvi image"""
    output = utils.getinfo(default_image_obj().ndvi)
    assert output['bands'][0]['id'] == 'ndvi'
    assert output['properties']['system:index'] == SCENE_ID
    assert output['properties']['system:time_start'] == SCENE_TIME
    assert output['properties']['image_id'] == COLL_ID + SCENE_ID


def test_Image_ndvi_constant_value(ndvi=0.8, expected=0.8, tol=0.0001):
    output = utils.constant_image_value(default_image_obj(ndvi=ndvi).ndvi)
    assert abs(output['ndvi'] - expected) <= tol


def test_Image_fc_properties():
    """Test if properties are set on the fc image"""
    output = utils.getinfo(default_image_obj().fc)
    assert output['bands'][0]['id'] == 'fc'
    assert output['properties']['system:index'] == SCENE_ID
    assert output['properties']['system:time_start'] == SCENE_TIME
    assert output['properties']['image_id'] == COLL_ID + SCENE_ID


def test_Image_crop_type_properties():
    """Test if properties are set on the crop type image"""
    output = utils.getinfo(default_image_obj().crop_type)
    assert output['bands'][0]['id'] == 'crop_type'
    # assert output['properties']['system:index'] == SCENE_ID
    # assert output['properties']['system:time_start'] == SCENE_TIME
    # assert output['properties']['image_id'] == COLL_ID + SCENE_ID


@pytest.mark.parametrize(
    'crop_type_source, xy, expected',
    [
        # Test spots around the Five Points CIMIS station
        ['USDA/NASS/CDL/2016', [-120.1130, 36.3360], 36],
        ['USDA/NASS/CDL/2016', [-120.1073, 36.3309], 69],
        ['USDA/NASS/CDL/2016', [-120.1080, 36.3459], 204],
        # Test a spot that has different CDL values through time
        ['USDA/NASS/CDL/2016', [-120.5953, 36.8721], 209],
        ['USDA/NASS/CDL/2017', [-120.5953, 36.8721], 24],
        ['USDA/NASS/CDL/2018', [-120.5953, 36.8721], 213],
        # Default image year is 2017 so value should match 2017 CDL
        ['USDA/NASS/CDL', [-120.5953, 36.8721], 24],
        ['projects/openet/assets/crop_type/v2021a', [-120.125, 36.3893], 47],
        ['projects/openet/assets/crop_type/v2023a', [-120.125, 36.3893], 47],
    ]
)
def test_Image_crop_type_point_value(crop_type_source, xy, expected):
    output = utils.point_image_value(default_image_obj(
        crop_type_source=crop_type_source).crop_type, xy)
    assert output['crop_type'] == expected


def test_Image_crop_class_properties():
    """Test if properties are set on the crop type image"""
    output = utils.getinfo(default_image_obj().crop_class)
    assert output['bands'][0]['id'] == 'crop_class'
    assert output['properties']['system:index'] == SCENE_ID
    assert output['properties']['system:time_start'] == SCENE_TIME
    assert output['properties']['image_id'] == COLL_ID + SCENE_ID


@pytest.mark.parametrize(
    'xy, expected',
    [
        # Test spots around the Five Points CIMIS station
        [[-120.113, 36.336], 1],
        [[-120.1073, 36.3309], 2],
        [[-120.108, 36.3459], 3],
    ]
)
def test_Image_crop_class_point_value(xy, expected):
    output = utils.point_image_value(default_image_obj(
        crop_type_source='USDA/NASS/CDL/2016').crop_class, xy)
    assert output['crop_class'] == expected


def test_Image_kc_properties():
    """Test if properties are set on the kc image"""
    output = utils.getinfo(default_image_obj().kc)
    assert output['bands'][0]['id'] == 'kc'
    assert output['properties']['system:index'] == SCENE_ID
    assert output['properties']['system:time_start'] == SCENE_TIME
    assert output['properties']['image_id'] == COLL_ID + SCENE_ID


def test_Image_et_fraction_properties():
    """Test if properties are set on the ET fraction image"""
    output = utils.getinfo(default_image_obj().et_fraction)
    assert output['bands'][0]['id'] == 'et_fraction'
    assert output['properties']['system:index'] == SCENE_ID
    assert output['properties']['system:time_start'] == SCENE_TIME
    assert output['properties']['image_id'] == COLL_ID + SCENE_ID


def test_Image_et_fraction_constant_value():
    # ET fraction method returns Kc
    output = utils.constant_image_value(default_image_obj(
        ndvi=0.8, crop_type_source=1).et_fraction)
    assert abs(output['et_fraction'] - 0.9859994736) <= 0.0001


def test_Image_et_reference_constant_value(et_reference=10.0, tol=0.0001):
    output = utils.constant_image_value(default_image_obj(
        et_reference_source=et_reference,
        et_reference_factor=0.85).et_reference)
    assert abs(output['et_reference'] - et_reference * 0.85) <= tol


def test_Image_et_reference_properties():
    """Test if properties are set on the reference ET image"""
    output = utils.getinfo(default_image_obj().et_reference)
    assert output['bands'][0]['id'] == 'et_reference'
    assert output['properties']['system:index'] == SCENE_ID
    assert output['properties']['system:time_start'] == SCENE_TIME
    assert output['properties']['image_id'] == COLL_ID + SCENE_ID


def test_Image_et_reference_source_exception():
    """Test that an Exception is raise for an invalid image ID"""
    with pytest.raises(Exception):
        utils.getinfo(default_image_obj(et_reference_source=None).et_reference)


# CGM - I'm not sure why this is commented out
# def test_Image_et_reference_band_exception():
#     """Test that an Exception is raise for an invalid et_reference band name"""
#     with pytest.raises(Exception):
#         utils.getinfo(default_image_obj(et_reference_band=None).et_reference)


def test_Image_et_properties():
    """Test if properties are set on the ET image"""
    output = utils.getinfo(default_image_obj().et)
    assert output['bands'][0]['id'] == 'et'
    assert output['properties']['system:index'] == SCENE_ID
    assert output['properties']['system:time_start'] == SCENE_TIME
    assert output['properties']['image_id'] == COLL_ID + SCENE_ID


def test_Image_et_constant_value():
    output = utils.constant_image_value(default_image_obj(
        et_reference_source=10, et_reference_factor=1.0, crop_type_source=1).et)
    assert abs(output['et'] - 10 * 0.986) <= 0.0001


def test_Image_mask_properties():
    """Test if properties are set on the time image"""
    output = utils.getinfo(default_image_obj().mask)
    assert output['bands'][0]['id'] == 'mask'
    assert output['properties']['system:index'] == SCENE_ID
    assert output['properties']['system:time_start'] == SCENE_TIME
    assert output['properties']['image_id'] == COLL_ID + SCENE_ID


def test_Image_mask_constant_value():
    output = utils.constant_image_value(default_image_obj(crop_type_source=1).mask)
    assert output['mask'] == 1


def test_Image_time_properties():
    """Test if properties are set on the time image"""
    output = utils.getinfo(default_image_obj().time)
    assert output['bands'][0]['id'] == 'time'
    assert output['properties']['system:index'] == SCENE_ID
    assert output['properties']['system:time_start'] == SCENE_TIME
    assert output['properties']['image_id'] == COLL_ID + SCENE_ID


def test_Image_time_constant_value():
    output = utils.constant_image_value(default_image_obj(crop_type_source=1).time)
    assert output['time'] == SCENE_TIME


def test_Image_calculate_variables_default():
    output = utils.getinfo(default_image_obj().calculate())
    assert set([x['id'] for x in output['bands']]) == set(['et'])


def test_Image_calculate_variables_custom():
    variables = ['ndvi']
    output = utils.getinfo(default_image_obj().calculate(variables))
    assert set([x['id'] for x in output['bands']]) == set(variables)


def test_Image_calculate_variables_all():
    variables = ['et', 'et_fraction', 'et_reference', 'fc', 'kc', 'mask', 'ndvi', 'time']
    output = utils.getinfo(default_image_obj().calculate(variables=variables))
    assert set([x['id'] for x in output['bands']]) == set(variables)


def test_Image_from_landsat_c2_sr_default_image():
    """Test that the classmethod is returning a class object"""
    output = sims.Image.from_landsat_c2_sr(input_image())
    assert type(output) == type(default_image_obj())


@pytest.mark.parametrize(
    'image_id',
    [
        # 'LANDSAT/LT04/C02/T1_L2/LT04_044033_19830812',
        'LANDSAT/LT05/C02/T1_L2/LT05_044033_20110716',
        'LANDSAT/LE07/C02/T1_L2/LE07_044033_20170708',
        'LANDSAT/LC08/C02/T1_L2/LC08_044033_20170716',
        'LANDSAT/LC09/C02/T1_L2/LC09_044033_20220127',
    ]
)
def test_Image_from_landsat_c2_sr_image_id(image_id):
    """Test instantiating the class from a Landsat image ID"""
    output = utils.getinfo(sims.Image.from_landsat_c2_sr(image_id).ndvi)
    assert output['properties']['system:index'] == image_id.split('/')[-1]


def test_Image_from_landsat_c2_sr_image():
    """Test instantiating the class from a Landsat ee.Image"""
    image_id = 'LANDSAT/LC08/C02/T1_L2/LC08_044033_20170716'
    output = utils.getinfo(sims.Image.from_landsat_c2_sr(ee.Image(image_id)).ndvi)
    assert output['properties']['system:index'] == image_id.split('/')[-1]


def test_Image_from_landsat_c2_sr_kc():
    """Test if ET fraction can be built from a Landsat images"""
    image_id = 'LANDSAT/LC08/C02/T1_L2/LC08_044033_20170716'
    output = utils.getinfo(sims.Image.from_landsat_c2_sr(image_id).kc)
    assert output['properties']['system:index'] == image_id.split('/')[-1]


def test_Image_from_landsat_c2_sr_et():
    """Test if ET can be built from a Landsat images"""
    image_id = 'LANDSAT/LC08/C02/T1_L2/LC08_044033_20170716'
    output = utils.getinfo(sims.Image.from_landsat_c2_sr(
        image_id, et_reference_source='IDAHO_EPSCOR/GRIDMET', et_reference_band='etr').et)
    assert output['properties']['system:index'] == image_id.split('/')[-1]


def test_Image_from_landsat_c2_sr_exception():
    """Test that an Exception is raise for an invalid image ID"""
    with pytest.raises(Exception):
        # Intentionally using .getInfo() since utils.getinfo() will catch the exception
        sims.Image.from_landsat_c2_sr(ee.Image('FOO')).ndvi.getInfo()


def test_Image_from_landsat_c2_sr_scaling():
    """Test if Landsat SR images images are being scaled"""
    sr_img = ee.Image('LANDSAT/LC08/C02/T1_L2/LC08_044033_20170716')
    # CGM - These reflectances should correspond to 0.1 for RED and 0.2 for NIR
    input_img = (
        ee.Image.constant([10909, 10909, 10909, 14545, 10909, 10909, 44177.6, 21824, 0])
        .rename(['SR_B2', 'SR_B3', 'SR_B4', 'SR_B5', 'SR_B6', 'SR_B7',
                 'ST_B10', 'QA_PIXEL', 'QA_RADSAT'])
        .set({'SPACECRAFT_ID': ee.String(sr_img.get('SPACECRAFT_ID')),
              'system:id': ee.String(sr_img.get('system:id')),
              'system:index': ee.String(sr_img.get('system:index')),
              'system:time_start': ee.Number(sr_img.get('system:time_start'))})
    )
    # cloud score masking and filter_flag option do not work with a constant image
    #   and must be explicitly set to False
    output = utils.constant_image_value(sims.Image.from_landsat_c2_sr(
        input_img, cloudmask_args={'cloud_score_flag': False, 'filter_flag': False}).ndvi)
    assert abs(output['ndvi'] - 0.333) <= 0.01


def test_Image_from_landsat_c2_sr_reflectance_type():
    """Test if reflectance_type property is being set"""
    image_id = 'LANDSAT/LC08/C02/T1_L2/LC08_044033_20170716'
    assert sims.Image.from_landsat_c2_sr(image_id).reflectance_type == 'SR'


def test_Image_from_landsat_c2_sr_cloud_mask_args():
    """Test if the cloud_mask_args parameter can be set (not if it works)"""
    output = sims.Image.from_landsat_c2_sr(
        'LANDSAT/LC08/C02/T1_L2/LC08_038031_20130828',
        cloudmask_args={'snow_flag': True, 'cirrus_flag': True})
    assert type(output) == type(default_image_obj())


def test_Image_from_landsat_c2_sr_cloud_score_mask_arg():
    """Test if the cloud_score_flag parameter can be set in cloudmask_args"""
    output = sims.Image.from_landsat_c2_sr(
        'LANDSAT/LC08/C02/T1_L2/LC08_044033_20170716',
        cloudmask_args={'cloud_score_flag': True, 'cloud_score_pct': 50})
    assert type(output) == type(default_image_obj())


@pytest.mark.parametrize(
    'image_id',
    [
        'LANDSAT/LC08/C02/T1_L2/LC08_044033_20170716',
    ]
)
def test_Image_from_image_id(image_id):
    """Test instantiating the class using the from_image_id method"""
    output = utils.getinfo(sims.Image.from_image_id(image_id).ndvi)
    assert output['properties']['system:index'] == image_id.split('/')[-1]
    assert output['properties']['image_id'] == image_id


def test_Image_from_method_kwargs():
    """Test that the init parameters can be passed through the helper methods"""
    assert sims.Image.from_landsat_c2_sr(
        'LANDSAT/LC08/C02/T1_L2/LC08_042035_20150713',
        et_reference_band='FOO').et_reference_band == 'FOO'



# import pprint
import re

import ee
import pytest

import openet.sims as sims
import openet.sims.utils as utils
# TODO: import utils from openet.core
# import openet.core.utils as utils

C02_COLLECTIONS = ['LANDSAT/LC08/C02/T1_L2', 'LANDSAT/LE07/C02/T1_L2']
# Image LE07_044033_20170724 is not (yet?) in LANDSAT/LE07/C02/T1_L2
C02_SCENE_ID_LIST = ['LC08_044033_20170716', 'LE07_044033_20170708']
START_DATE = '2017-07-01'
END_DATE = '2017-08-01'
SCENE_GEOM = (-121.91, 38.99, -121.89, 39.01)
SCENE_POINT = (-121.9, 39)
VARIABLES = {'et', 'et_fraction', 'et_reference'}
TEST_POINT = (-121.5265, 38.7399)


def default_coll_args():
    return {
        'collections': C02_COLLECTIONS,
        'geometry': ee.Geometry.Point(SCENE_POINT),
        'start_date': START_DATE,
        'end_date': END_DATE,
        'variables': list(VARIABLES),
        'cloud_cover_max': 70,
        'et_reference_source': 'IDAHO_EPSCOR/GRIDMET',
        'et_reference_band': 'eto',
        'et_reference_factor': 0.85,
        'et_reference_resample': 'nearest',
        'model_args': {
            'cloudmask_args': {'cloud_score_flag': False, 'filter_flag': False},
        },
        'filter_args': {},
    }


def default_coll_obj(**kwargs):
    args = default_coll_args().copy()
    args.update(kwargs)
    return sims.Collection(**args)


def parse_scene_id(output_info):
    output = [x['properties']['system:index'] for x in output_info['features']]
    # Strip merge indices (this works for Landsat and Sentinel image IDs
    return sorted(['_'.join(x.split('_')[-3:]) for x in output])


def test_Collection_init_default_parameters():
    """Test if init sets default parameters"""
    args = default_coll_args().copy()
    # These values are being set above but have defaults that need to be checked
    del args['et_reference_source']
    del args['et_reference_band']
    del args['et_reference_factor']
    del args['et_reference_resample']
    del args['variables']

    m = sims.Collection(**args)
    assert m.variables is None
    assert m.et_reference_source is None
    assert m.et_reference_band is None
    assert m.et_reference_factor is None
    assert m.et_reference_resample is None
    assert m.cloud_cover_max == 70
    # assert m.model_args == {}
    assert m.filter_args == {}
    assert set(m._interp_vars) == {'ndvi', 'et_fraction'}


def test_Collection_init_collection_str(coll_id='LANDSAT/LC08/C02/T1_L2'):
    """Test if a single coll_id str is converted to a single item list"""
    assert default_coll_obj(collections=coll_id).collections == [coll_id]


def test_Collection_init_cloud_cover_max_str():
    """Test if cloud_cover_max strings are converted to float"""
    assert default_coll_obj(cloud_cover_max='70').cloud_cover_max == 70


@pytest.mark.parametrize(
    'coll_id, start_date, end_date',
    [
        ['LANDSAT/LT04/C02/T1_L2', '1981-01-01', '1982-01-01'],
        ['LANDSAT/LT04/C02/T1_L2', '1994-01-01', '1995-01-01'],
        ['LANDSAT/LT05/C02/T1_L2', '1983-01-01', '1984-01-01'],
        ['LANDSAT/LT05/C02/T1_L2', '2012-01-01', '2013-01-01'],
        ['LANDSAT/LE07/C02/T1_L2', '1998-01-01', '1999-01-01'],
        ['LANDSAT/LE07/C02/T1_L2', '2022-01-01', '2023-01-01'],
        ['LANDSAT/LC08/C02/T1_L2', '2012-01-01', '2013-01-01'],
        ['LANDSAT/LC09/C02/T1_L2', '2021-01-01', '2022-01-01'],
    ]
)
def test_Collection_init_collection_filter(coll_id, start_date, end_date):
    """Test that collection IDs are filtered based on start/end dates"""
    # The target collection ID should be removed from the collections lists
    assert default_coll_obj(collections=coll_id, start_date=start_date,
                            end_date=end_date).collections == []


def test_Collection_init_startdate_exception():
    """Test if Exception is raised for invalid start date formats"""
    with pytest.raises(ValueError):
        default_coll_obj(start_date='1/1/2000', end_date='2000-01-02')


def test_Collection_init_enddate_exception():
    """Test if Exception is raised for invalid end date formats"""
    with pytest.raises(ValueError):
        default_coll_obj(start_date='2000-01-01', end_date='1/2/2000')


def test_Collection_init_swapped_date_exception():
    """Test if Exception is raised when start_date == end_date"""
    with pytest.raises(ValueError):
        default_coll_obj(start_date='2017-01-01', end_date='2017-01-01')


def test_Collection_init_invalid_collections_exception():
    """Test if Exception is raised for an invalid collection ID"""
    with pytest.raises(ValueError):
        default_coll_obj(collections=['FOO'])


def test_Collection_init_cloud_cover_exception():
    """Test if Exception is raised for an invalid cloud_cover_max"""
    with pytest.raises(TypeError):
        default_coll_obj(cloud_cover_max='A')
    with pytest.raises(ValueError):
        default_coll_obj(cloud_cover_max=-1)
    with pytest.raises(ValueError):
        default_coll_obj(cloud_cover_max=101)


def test_Collection_build_default():
    output = utils.getinfo(default_coll_obj()._build())
    assert output['type'] == 'ImageCollection'
    assert parse_scene_id(output) == C02_SCENE_ID_LIST
    # Check that the variables being set in the default collection object are returned
    assert {y['id'] for x in output['features'] for y in x['bands']} == VARIABLES


def test_Collection_build_variables_custom(variable='ndvi'):
    # Check that setting the build variables overrides the collection variables
    output = utils.getinfo(
        default_coll_obj()._build(variables=[variable]).first().bandNames()
    )
    assert set(output) == {variable}


def test_Collection_build_variables_none():
    """Test for exception if variables is set to None in method call"""
    with pytest.raises(ValueError):
        utils.getinfo(default_coll_obj(variables=None)._build(variables=None))


def test_Collection_build_variables_not_set():
    """Test for exception if variables is not set in method since default is None"""
    with pytest.raises(ValueError):
        utils.getinfo(default_coll_obj(variables=None)._build())


def test_Collection_build_variables_empty_list():
    # Setting variables to an empty list should return the merged Landsat collection
    output = utils.getinfo(
        default_coll_obj(collections=C02_COLLECTIONS, variables=None)
        ._build(variables=[]).first().bandNames()
    )
    assert 'SR_B3' in output


def test_Collection_build_invalid_variable_exception():
    """Test if Exception is raised for an invalid variable"""
    with pytest.raises(ValueError):
        utils.getinfo(default_coll_obj()._build(variables=['FOO']))


def test_Collection_build_dates():
    """Check that dates passed to build function override Class dates"""
    coll_obj = default_coll_obj(start_date='2017-08-01', end_date='2017-09-01')
    output = utils.getinfo(coll_obj._build(start_date='2017-07-16', end_date='2017-07-17'))
    assert parse_scene_id(output) == ['LC08_044033_20170716']


def test_Collection_build_landsat_c2_sr():
    """Test if the Landsat SR collections can be built"""
    coll_obj = default_coll_obj(collections=['LANDSAT/LC08/C02/T1_L2', 'LANDSAT/LE07/C02/T1_L2'])
    output = utils.getinfo(coll_obj._build())
    assert parse_scene_id(output) == C02_SCENE_ID_LIST
    assert {y['id'] for x in output['features'] for y in x['bands']} == VARIABLES


def test_Collection_build_exclusive_enddate():
    """Test if the end_date is exclusive"""
    output = utils.getinfo(default_coll_obj(end_date='2017-07-24')._build())
    assert [x for x in parse_scene_id(output) if int(x[-8:]) >= 20170724] == []


def test_Collection_build_cloud_cover():
    """Test if the cloud cover max parameter is being applied"""
    # CGM - The filtered images should probably be looked up programmatically
    output = utils.getinfo(default_coll_obj(cloud_cover_max=0.5)._build(variables=['et']))
    assert 'LE07_044033_20170724' not in parse_scene_id(output)


@pytest.mark.parametrize(
    'collection, start_date, end_date',
    [
        ['LANDSAT/LT05/C02/T1_L2', '2012-01-01', '2013-01-01'],
    ]
)
def test_Collection_build_filter_dates_lt05(collection, start_date, end_date):
    """Test that bad Landsat 5 images are filtered"""
    coll = default_coll_obj(
        collections=[collection], start_date=start_date, end_date=end_date,
        geometry=ee.Geometry.Rectangle(-125, 25, -65, 50),
    )
    output = utils.getinfo(coll._build(variables=['et']))
    assert parse_scene_id(output) == []


@pytest.mark.parametrize(
    'collection, start_date, end_date',
    [
        ['LANDSAT/LE07/C02/T1_L2', '2022-01-01', '2023-01-01'],
    ]
)
def test_Collection_build_filter_dates_le07(collection, start_date, end_date):
    """Test that Landsat 7 images after 2021 are filtered"""
    coll = default_coll_obj(
        collections=[collection], start_date=start_date, end_date=end_date,
        geometry=ee.Geometry.Rectangle(-125, 25, -65, 50),
    )
    output = utils.getinfo(coll._build(variables=['et']))
    assert parse_scene_id(output) == []


@pytest.mark.parametrize(
    'collection, start_date, end_date',
    [
        ['LANDSAT/LC08/C02/T1_L2', '2013-01-01', '2013-04-01'],
    ]
)
def test_Collection_build_filter_dates_lc08(collection, start_date, end_date):
    """Test that pre-op Landsat 8 images before 2013-04-01 are filtered"""
    coll = default_coll_obj(
        collections=[collection], start_date=start_date, end_date=end_date,
        geometry=ee.Geometry.Rectangle(-125, 25, -65, 50),
    )
    output = utils.getinfo(coll._build(variables=['et']))
    assert not [x for x in parse_scene_id(output)
                if x.split('_')[-1] < end_date.replace('-', '')]
    assert parse_scene_id(output) == []


@pytest.mark.parametrize(
    'collection, start_date, end_date',
    [
        ['LANDSAT/LC09/C02/T1_L2', '2021-11-01', '2022-01-01'],
    ]
)
def test_Collection_build_filter_dates_lc09(collection, start_date, end_date):
    """Test that Landsat 9 images before 2022-01-01 are filtered"""
    coll = default_coll_obj(
        collections=[collection], start_date=start_date, end_date=end_date,
        geometry=ee.Geometry.Rectangle(-125, 25, -65, 50),
    )
    output = utils.getinfo(coll._build(variables=['et']))
    assert not [x for x in parse_scene_id(output)
                if x.split('_')[-1] < end_date.replace('-', '')]
    assert parse_scene_id(output) == []


def test_Collection_build_filter_args_keyword():
    # Need to test with two collections to catch bug when deepcopy isn't used
    collections = ['LANDSAT/LC08/C02/T1_L2', 'LANDSAT/LE07/C02/T1_L2']
    wrs2_filter = [
        {'type': 'equals', 'leftField': 'WRS_PATH', 'rightValue': 44},
        {'type': 'equals', 'leftField': 'WRS_ROW', 'rightValue': 33},
    ]
    coll_obj = default_coll_obj(
        collections=collections,
        geometry=ee.Geometry.Rectangle(-125, 35, -120, 40),
        filter_args={c: wrs2_filter for c in collections},
    )
    output = utils.getinfo(coll_obj._build(variables=['et']))
    assert {x[5:11] for x in parse_scene_id(output)} == {'044033'}


def test_Collection_build_filter_args_eeobject():
    # Need to test with two collections to catch bug when deepcopy isn't used
    collections = ['LANDSAT/LC08/C02/T1_L2', 'LANDSAT/LE07/C02/T1_L2']
    wrs2_filter = ee.Filter.And(
        ee.Filter.equals('WRS_PATH', 44), ee.Filter.equals('WRS_ROW', 33)
    )
    coll_obj = default_coll_obj(
        collections=collections,
        geometry=ee.Geometry.Rectangle(-125, 35, -120, 40),
        filter_args={c: wrs2_filter for c in collections},
    )
    output = utils.getinfo(coll_obj._build(variables=['et']))
    assert {x[5:11] for x in parse_scene_id(output)} == {'044033'}


def test_Collection_overpass_default():
    """Test overpass method with default values (variables from Class init)"""
    output = utils.getinfo(default_coll_obj().overpass())
    assert {y['id'] for x in output['features'] for y in x['bands']} == VARIABLES
    assert parse_scene_id(output) == C02_SCENE_ID_LIST


def test_Collection_overpass_class_variables():
    """Test that custom class variables are passed through to build function"""
    output = utils.getinfo(default_coll_obj(variables=['et']).overpass())
    assert {y['id'] for x in output['features'] for y in x['bands']} == {'et'}


def test_Collection_overpass_method_variables():
    """Test that custom method variables are passed through to build function"""
    output = utils.getinfo(default_coll_obj().overpass(variables=['et']))
    assert {y['id'] for x in output['features'] for y in x['bands']} == {'et'}


def test_Collection_overpass_no_variables_exception():
    """Test if Exception is raised if variables is not set in init or method"""
    with pytest.raises(ValueError):
        utils.getinfo(default_coll_obj(variables=[]).overpass())


def test_Collection_interpolate_default():
    """Default t_interval should be custom"""
    output = utils.getinfo(default_coll_obj().interpolate())
    assert output['type'] == 'ImageCollection'
    assert parse_scene_id(output) == ['20170701']
    assert {y['id'] for x in output['features'] for y in x['bands']} == VARIABLES


@pytest.mark.parametrize('use_joins', [True, False])
def test_Collection_interpolate_use_joins(use_joins):
    """Only checking if the parameter is accepted and runs for now"""
    output = utils.getinfo(default_coll_obj().interpolate(use_joins=use_joins))
    assert output['type'] == 'ImageCollection'
    assert parse_scene_id(output) == ['20170701']


def test_Collection_interpolate_variables_custom_et():
    output = utils.getinfo(default_coll_obj().interpolate(variables=['et']))
    assert {y['id'] for x in output['features'] for y in x['bands']} == {'et'}


def test_Collection_interpolate_variables_custom_ndvi():
    output = utils.getinfo(default_coll_obj().interpolate(variables=['ndvi']))
    assert {y['id'] for x in output['features'] for y in x['bands']} == {'ndvi'}


# DEADBEEF
# def test_Collection_interpolate_variables_custom_daily_count():
#     output = utils.getinfo(default_coll_obj().interpolate(variables=['daily_count']))
#     assert {y['id'] for x in output['features'] for y in x['bands']} == {'daily_count'}


def test_Collection_interpolate_t_interval_daily():
    """Test if the daily time interval parameter works

    Since end_date is exclusive last image date will be one day earlier
    """
    coll_obj = default_coll_obj(start_date='2017-07-01', end_date='2017-07-05')
    output = utils.getinfo(coll_obj.interpolate(t_interval='daily'))
    assert output['type'] == 'ImageCollection'
    assert parse_scene_id(output)[0] == '20170701'
    assert parse_scene_id(output)[-1] == '20170704'
    assert {y['id'] for x in output['features'] for y in x['bands']} == VARIABLES


def test_Collection_interpolate_t_interval_monthly():
    """Test if the monthly time interval parameter works"""
    output = utils.getinfo(default_coll_obj().interpolate(t_interval='monthly'))
    assert output['type'] == 'ImageCollection'
    assert parse_scene_id(output) == ['201707']
    assert {y['id'] for x in output['features'] for y in x['bands']} == VARIABLES


def test_Collection_interpolate_t_interval_custom():
    """Test if the custom time interval parameter works"""
    output = utils.getinfo(default_coll_obj().interpolate(t_interval='custom'))
    assert output['type'] == 'ImageCollection'
    assert parse_scene_id(output) == ['20170701']
    assert {y['id'] for x in output['features'] for y in x['bands']} == VARIABLES


# TODO: Write test for monthly interpolation with a date range that is too short


# def test_Collection_interpolate_interp_days():
#     """Test if the interpolate interp_days parameter works"""
#     # Is there any way to test this without pulling values at a point?


# NOTE: For the following tests the collection class is not being
#   re-instantiated for each test so it is necessary to clear the model_args
def test_Collection_interpolate_et_reference_source_not_set():
    """Test if Exception is raised if et_reference_source is not set"""
    with pytest.raises(ValueError):
        utils.getinfo(default_coll_obj(et_reference_source=None, model_args={}).interpolate())


def test_Collection_interpolate_et_reference_band_not_set():
    """Test if Exception is raised if et_reference_band is not set"""
    with pytest.raises(ValueError):
        utils.getinfo(default_coll_obj(et_reference_band=None, model_args={}).interpolate())


def test_Collection_interpolate_et_reference_factor_exception():
    """Test if Exception is raised if et_reference_factor is not a number or negative"""
    with pytest.raises(ValueError):
        utils.getinfo(default_coll_obj(et_reference_factor=-1, model_args={}).interpolate())


def test_Collection_interpolate_et_reference_resample_exception():
    """Test if Exception is raised if et_reference_resample is not set"""
    with pytest.raises(ValueError):
        utils.getinfo(
            default_coll_obj(et_reference_resample='deadbeef', model_args={}).interpolate()
        )


def test_Collection_interpolate_et_reference_params_kwargs():
    """Test setting et_reference parameters in the Collection init args"""
    coll = default_coll_obj(
        et_reference_source='IDAHO_EPSCOR/GRIDMET', et_reference_band='eto',
        et_reference_factor=0.5, et_reference_resample='bicubic', model_args={},
    )
    output = utils.getinfo(coll.interpolate(use_joins=True))
    assert {y['id'] for x in output['features'] for y in x['bands']} == VARIABLES
    assert output['features'][0]['properties']['et_reference_factor'] == 0.5
    assert output['features'][0]['properties']['et_reference_resample'] == 'bicubic'


def test_Collection_interpolate_et_reference_params_model_args():
    """Test setting et_reference parameters in the model_args"""
    coll = default_coll_obj(
        et_reference_source=None, et_reference_band=None,
        et_reference_factor=None, et_reference_resample=None,
        model_args={'et_reference_source': 'IDAHO_EPSCOR/GRIDMET',
                    'et_reference_band': 'eto',
                    'et_reference_factor': 0.5,
                    'et_reference_resample': 'bicubic'},
    )
    output = utils.getinfo(coll.interpolate(use_joins=True))
    assert {y['id'] for x in output['features'] for y in x['bands']} == VARIABLES
    assert output['features'][0]['properties']['et_reference_factor'] == 0.5
    assert output['features'][0]['properties']['et_reference_resample'] == 'bicubic'


def test_Collection_interpolate_et_reference_params_interpolate_args():
    """Test setting et_reference parameters in the interpolate call"""
    et_reference_args = {'et_reference_source': 'IDAHO_EPSCOR/GRIDMET',
                         'et_reference_band': 'eto',
                         'et_reference_factor': 0.5,
                         'et_reference_resample': 'bicubic'}
    coll = default_coll_obj(
        et_reference_source=None, et_reference_band=None,
        et_reference_factor=None, et_reference_resample=None, model_args={},
    )
    output = utils.getinfo(coll.interpolate(use_joins=True, **et_reference_args))
    assert {y['id'] for x in output['features'] for y in x['bands']} == VARIABLES
    assert output['features'][0]['properties']['et_reference_factor'] == 0.5
    assert output['features'][0]['properties']['et_reference_resample'] == 'bicubic'


def test_Collection_interpolate_t_interval_exception():
    """Test if Exception is raised for an invalid t_interval parameter"""
    with pytest.raises(ValueError):
        utils.getinfo(default_coll_obj().interpolate(t_interval='DEADBEEF'))


def test_Collection_interpolate_interp_method_exception():
    """Test if Exception is raised for an invalid interp_method parameter"""
    with pytest.raises(ValueError):
        utils.getinfo(default_coll_obj().interpolate(interp_method='DEADBEEF'))


def test_Collection_interpolate_interp_days_exception():
    """Test if Exception is raised for an invalid interp_days parameter"""
    with pytest.raises(ValueError):
        utils.getinfo(default_coll_obj().interpolate(interp_days=0))


def test_Collection_interpolate_no_variables_exception():
    """Test if Exception is raised if variables is not set in init or method"""
    with pytest.raises(ValueError):
        utils.getinfo(default_coll_obj(variables=[]).interpolate())


# def test_Collection_overpass_values():
#     output = utils.point_coll_value(
#         default_coll_obj().overpass(), xy=TEST_POINT, scale=30)
#     pprint.pprint(output)
#     assert False


# def test_Collection_interpolate_values_daily():
#     output = utils.point_coll_value(
#         default_coll_obj().interpolate(t_interval='daily'),
#         xy=TEST_POINT, scale=30)
#     pprint.pprint(output)
#     assert False


# def test_Collection_interpolate_values_monthly():
#     output = utils.point_coll_value(
#         default_coll_obj().interpolate(t_interval='monthly'),
#         xy=TEST_POINT, scale=30)
#     pprint.pprint(output)
#     assert False


# def test_Collection_interpolate_values_custom():
#     output = utils.point_coll_value(
#         default_coll_obj().interpolate(t_interval='custom'),
#         xy=TEST_POINT, scale=30)
#     pprint.pprint(output)
#     assert False


# def test_Collection_interpolate_values_interp_days():
#     args = default_coll_args()
#     args['start_date'] = '2017-06-30'
#     output = utils.point_coll_value(
#         sims.Collection(**args).overpass(
#             variables=['kc', 'et_reference']), xy=TEST_POINT, scale=30)
#     pprint.pprint(output)
#
#     output = utils.point_coll_value(
#         sims.Collection(**args).interpolate(
#             variables=['kc', 'et_reference'], t_interval='daily',
#             interp_days=32),
#         xy=TEST_POINT, scale=30)
#     pprint.pprint(output)
#     assert False


def test_Collection_interpolate_output_type_default():
    """Test if output_type parameter is defaulting to float"""
    test_vars = ['et', 'et_reference', 'et_fraction', 'ndvi', 'count']
    output = utils.getinfo(default_coll_obj(variables=test_vars).interpolate())
    output = output['features'][0]['bands']
    bands = {info['id']: i for i, info in enumerate(output)}
    assert output[bands['et']]['data_type']['precision'] == 'float'
    assert output[bands['et_reference']]['data_type']['precision'] == 'float'
    assert output[bands['et_fraction']]['data_type']['precision'] == 'float'
    assert output[bands['ndvi']]['data_type']['precision'] == 'float'
    assert output[bands['count']]['data_type']['precision'] == 'int'


def test_Collection_interpolate_custom_model_args():
    """Test passing in a model specific parameter through model_args"""
    model_args = {'crop_type_source': 'projects/openet/crop_type/annual_provisional'}
    output = utils.getinfo(default_coll_obj(model_args=model_args).interpolate(use_joins=True))
    output = output['features'][0]['properties']
    assert output['crop_type_source'] == 'projects/openet/crop_type/annual_provisional'


def test_Collection_interpolate_only_interpolate_images():
    """Test if count band is returned if no images in the date range"""
    variables = {'et', 'count'}
    coll = default_coll_obj(
        collections=['LANDSAT/LC08/C02/T1_L2'],
        geometry=ee.Geometry.Point(-123.623, 44.745),
        start_date='2017-04-01', end_date='2017-04-30',
        variables=list(variables), cloud_cover_max=70,
    )
    output = utils.getinfo(coll.interpolate())
    assert {y['id'] for x in output['features'] for y in x['bands']} == variables


def test_Collection_interpolate_model_properties():
    """Check that setting the model name and version from importlib.metadata works"""
    output = utils.getinfo(default_coll_obj().interpolate().first())
    version_re = re.compile(r'[0-9]+[.][0-9]+[.][0-9]+\w*')
    assert output['properties']['model_name'] == 'openet-sims'
    assert version_re.match(output['properties']['model_version'])


@pytest.mark.parametrize(
    'collections, scene_id_list',
    [
        [['LANDSAT/LC08/C02/T1_L2', 'LANDSAT/LE07/C02/T1_L2'], C02_SCENE_ID_LIST],
    ]
)
def test_Collection_get_image_ids(collections, scene_id_list):
    # get_image_ids method makes a getInfo call internally
    output = default_coll_obj(collections=collections, variables=None).get_image_ids()
    assert type(output) is list
    assert set(x.split('/')[-1] for x in output) == set(scene_id_list)



# import pprint

import ee

from . import data
from . import utils


# def lazy_property(fn):
#     """Decorator that makes a property lazy-evaluated
#
#     https://stevenloria.com/lazy-properties/
#     """
#     attr_name = '_lazy_' + fn.__name__
#
#     @property
#     def _lazy_property(self):
#         if not hasattr(self, attr_name):
#             setattr(self, attr_name, fn(self))
#         return getattr(self, attr_name)
#     return _lazy_property


class Model():
    """GEE based model for computing SIMS ETcb"""

    def __init__(
        self,
        # CGM - Switch to ee.Date or time_start instead?
        year,
        doy,
        crop_type_source='USDA/NASS/CDL',
        crop_type_remap='CDL',
        crop_type_kc_flag=False,
        crop_type_annual_skip_flag=False,
        mask_non_ag_flag=True,
        water_kc_flag=True,
        reflectance_type='SR',
    ):
        """Earth Engine based SIMS model object

        Parameters
        ----------
        year : ee.Number
        doy : ee.Number
            Day of year
        crop_type_source : str, optional
            Crop type source.  The default is the Cropland Data Layer (CDL) assets.
            The source should be an Earth Engine Image ID (or ee.Image).
            Currently only the OpenET collection and CDL images are supported.
        crop_type_remap : {'CDL'}, optional
            Currently only CDL crop type values are supported.
        crop_type_kc_flag : bool, optional
            If True, compute Kc using crop type specific coefficients.
            If False, use generic crop class coefficients. The default is False.
        crop_type_annual_skip_flag : bool, optional
            If True, the crop type specific coefficients are NOT used for annual crops.
            If False, the crop type specific coefficients are used for annual crops.
            This flag is only applied/used if crop_type_kc_flag is also True.
            The default is False.
        mask_non_ag_flag : bool, optional
            If True, mask all pixels that don't map to a crop_class.
            The default is False.
        water_kc_flag : bool, optional
            If True, set Kc for water pixels to 1.05.  The default is True.
        reflectance_type : {'SR', 'TOA'}, optional
            Used to select the fractional cover equation (the default is 'SR').

        """

        # Model could inherit these values from Image instead of being passed in
        self.year = year
        self.doy = doy
        # self.date = ee.Date(self.time_start)
        # self.year = ee.Number(self.date.get('year'))
        # self.doy = self.date.getRelative('day', 'year').add(1).int()

        self.crop_type_source = crop_type_source
        self.crop_type_remap = crop_type_remap
        self.crop_type_kc_flag = crop_type_kc_flag
        self.crop_type_annual_skip_flag = crop_type_annual_skip_flag
        self.mask_non_ag_flag = mask_non_ag_flag
        self.water_kc_flag = water_kc_flag

        # CGM - Trying out setting these as properties in init
        #   instead of as lazy properties below
        self.crop_data = self._crop_data()
        self.crop_type = self._crop_type()
        self.crop_class = crop_data_image('crop_class', self.crop_type, self.crop_data, 0)

        # Manually set the crop data parameter images as class properties
        # Set default values for some properties to ensure fr == 1
        self.h_max = crop_data_image('h_max', self.crop_type, self.crop_data)
        self.m_l = crop_data_image('m_l', self.crop_type, self.crop_data)
        self.fr_mid = crop_data_image('fr_mid', self.crop_type, self.crop_data, 1)
        self.fr_end = crop_data_image('fr_end', self.crop_type, self.crop_data, 1)
        self.ls_start = crop_data_image('ls_start', self.crop_type, self.crop_data, 1)
        self.ls_stop = crop_data_image('ls_stop', self.crop_type, self.crop_data, 365)
        # setattr('h_max', crop_data_image(
        #     'h_max', self.crop_type, self.crop_data))

        self.reflectance_type = reflectance_type
        # TODO: Should type be checked (and exception raised) here or in fc()?
        #   Being raised in fc() for now since it is not a lazy property
        # if self.reflectance_type not in ['SR', 'TOA']:
        #     raise ValueError(f'unsupported reflectance type: {reflectance_type}')

    # CGM - It would be nice if kc and fc were lazy properties but then fc and
    #   ndvi would need to part of self (inherited from Image?).
    # @lazy_property
    def kc(self, ndvi):
        """Crop coefficient (kc) for all crop classes and types

        Parameters
        ----------
        ndvi : ee.Image
            Normalized difference vegetation index.

        Returns
        -------
        ee.Image

        Notes
        ----
        Generic Fc-Kcb conversion for:
        Annuals:
        Melton, F., L. Johnson, C. Lund, L. Pierce, A. Michaelis, S. Hiatt,
            A. Guzman, D. Adhikari, A. Purdy, C. Rosevelt, P. Votava, T. Trout,
            B. Temesgen, K. Frame, E. Sheffner, and R. Nemani (2012).
            Satellite Irrigation Management Support with the Terrestrial
            Observation and Prediction System: An Operational Framework for
            Integration of Satellite and Surface Observations to Support
            Improvements in Agricultural Water Resource Management.
            IEEE J. Selected Topics in Applied Earth Observations & Remote Sensing
            5:1709-1721.  [FIG 2b]

        Crop specific Fc-Kcb conversion:
        Allen, R., and L. Pereira (2009).  Estimating crop coefficients from
            fraction of ground cover and height.  Irrigation Science 28:17-34.
            DOI 10.1007/s00271-009-0182-z
            [EQNS 10 (Kd); 7a (Kcb_full) using tree/vine Fr vals from Table 2; 5a (Kcb)]

        """
        fc = self.fc(ndvi)

        # Start with the generic NDVI-Kc relationship to initialize Kc
        kc = self.kc_generic(ndvi)

        # Apply generic crop class Kc functions
        kc = kc.where(self.crop_class.eq(1), self.kc_row_crop(fc))
        kc = kc.where(self.crop_class.eq(2), self._kcb(self._kd_vine(fc)).clamp(0, 1.1))
        kc = kc.where(self.crop_class.eq(3), self.kc_tree(fc))
        kc = kc.where(self.crop_class.eq(5), self.kc_rice(fc, ndvi))
        kc = kc.where(self.crop_class.eq(6), self.kc_fallow(fc, ndvi))
        kc = kc.where(self.crop_class.eq(7), self.kc_grass_pasture(fc, ndvi))

        if self.crop_type_kc_flag:
            # Apply crop type specific Kc functions
            # h_max.gte(0) is needed to select pixels that have custom
            #   coefficient values in the crop_data dictionary
            # The h_max image was built with all non-remapped crop_types as nodata
            if not self.crop_type_annual_skip_flag:
                kc = kc.where(self.crop_class.eq(1).And(self.h_max.gte(0)),
                              self._kcb(self._kd_row_crop(fc)))

            kc = kc.where(self.crop_class.eq(3).And(self.h_max.gte(0)),
                          self._kcb(self._kd_tree(fc)).clamp(0, 1.2))

            # CGM - Commenting out for now
            # kc = kc.where(
            #     self.crop_class.eq(3).And(self.h_max.gte(0)).And(kc.gte(0.2)),
            #     self._kcb(self._kd_tree(fc), kc_min=0.5).clamp(0, 1.2))

        # CGM - Is it okay to apply this after all the other Kc functions?
        #   Should we only apply this to non-ag crop classes?
        if self.water_kc_flag:
            kc = kc.where(ndvi.lt(0).And(self.crop_class.eq(0)), 1.05)
            # kc = kc.where(ndvi.lt(0), 1.05)

        if self.mask_non_ag_flag:
            kc = kc.updateMask(self.crop_class.gt(0))

        return kc.rename(['kc'])

    # @lazy_property
    def fc(self, ndvi):
        """Fraction of cover (fc)

        Parameters
        ----------
        ndvi : ee.Image

        Returns
        -------
        ee.Image

        Raises
        ------
        ValueError if reflectance type is not supported

        References
        ----------


        """
        if self.reflectance_type == 'SR':
            fc = ndvi.multiply(1.26).subtract(0.18)
        elif self.reflectance_type == 'TOA':
            fc = ndvi.multiply(1.465).subtract(0.139)
        else:
            raise ValueError(f'Unsupported reflectance type: {self.reflectance_type}')

        return fc.clamp(0, 1).rename(['fc'])

    def _crop_type(self):
        """Crop type

        Parameters
        ----------
        crop_type_source : int, str
            CDL image collection ID: 'USDA/NASS/CDL'
                Collection will be filtered to a single year that is closest
                to the Image year.
            CDL image ID for a specific year: 'USDA/NASS/CDL/2018'
            OpenET crop type image collection ID:
                'projects/openet/assets/crop_type/v2023a'
                Collection will be mosaiced to a single image.
            Integer (will be converted to an EE constant image)
        year : ee.Number
            Year is needed if crop_type_source is the CDL collection.

        Returns
        -------
        ee.Image

        Raises
        ------
        ValueError for unsupported crop_type_sources

        """
        properties = ee.Dictionary()

        if utils.is_number(self.crop_type_source):
            # Interpret numbers as constant images
            # CGM - Should we use the ee_types here instead?
            #   i.e. ee.ee_types.isNumber(self.et_reference_source)
            crop_type_img = ee.Image.constant(self.crop_type_source)
            #     .rename('crop_type')
            # properties = properties.set('id', 'constant')
        elif (type(self.crop_type_source) is str and
              self.crop_type_source.upper() == 'USDA/NASS/CDL'):
            # Assume source is the CDL image collection ID
            # Use the CDL image closest to the image date
            # Don't use CDL images before 2008
            cdl_coll = ee.ImageCollection('USDA/NASS/CDL')
            cdl_year = (
                ee.Number(self.year)
                .max(2008)
                .min(ee.Date(cdl_coll.aggregate_max('system:time_start')).get('year'))
            )
            cdl_coll = (
                cdl_coll
                .filterDate(ee.Date.fromYMD(cdl_year, 1, 1),
                            ee.Date.fromYMD(cdl_year.add(1), 1, 1))
                .select(['cropland'])
            )
            crop_type_img = ee.Image(cdl_coll.first())
            properties = properties.set('id', crop_type_img.get('system:id'))
        elif (type(self.crop_type_source) is str and
                self.crop_type_source.upper().startswith('USDA/NASS/CDL')):
            # Assume source is a single CDL image ID
            crop_type_img = ee.Image(self.crop_type_source).select(['cropland'])
            properties = properties.set('id', crop_type_img.get('system:id'))
        elif (type(self.crop_type_source) is str and
              ('projects/openet/crop_type' in self.crop_type_source.lower() or
               'projects/openet/assets/crop_type' in self.crop_type_source.lower())):
            # Assume source is an OpenET crop type image collection ID
            # Use the crop type image closest to the image date
            crop_coll = ee.ImageCollection(self.crop_type_source)
            cdl_year = (
                ee.Number(self.year)
                .max(ee.Date(crop_coll.aggregate_min('system:time_start')).get('year'))
                .min(ee.Date(crop_coll.aggregate_max('system:time_start')).get('year'))
            )
            crop_type_coll = (
                ee.ImageCollection(self.crop_type_source)
                .filterDate(ee.Date.fromYMD(cdl_year, 1, 1),
                            ee.Date.fromYMD(cdl_year.add(1), 1, 1))
            )
            crop_type_img = crop_type_coll.mosaic()
            properties = properties.set('id', crop_type_coll.get('system:id'))
        # TODO: Support ee.Image and ee.ImageCollection sources
        # elif isinstance(self.crop_type_source, computedobject.ComputedObject):
        else:
            raise ValueError(f'unsupported crop_type_source: {self.crop_type_source}')

        # Should the image properties be set onto the image also?
        return crop_type_img.rename(['crop_type']).set(properties)

    def _crop_data(self):
        """Load the crop data dictionary

        Currently the data is in data.py but this method could also be used to
        read the data from a json or csv file.

        Parameters
        ----------
        crop_type : ee.Image
        crop_type_remap : str

        Returns
        -------
        dict

        Raises
        ------
        ValueError for unsupported crop_type_remap

        """
        if self.crop_type_remap.upper() == 'CDL':
            return data.cdl
        else:
            raise ValueError(f'unsupported crop_type_remap: "{self.crop_type_remap}"')

    def kc_generic(self, ndvi):
        """Generic crop coefficient based on linear function of NDVI

        Parameters
        ----------
        ndvi : ee.Image
            Normalized difference vegetation index.

        Returns
        -------
        ee.Image

        """
        return ndvi.multiply(1.25).add(0.2).max(0).rename(['kc'])

    def kc_row_crop(self, fc):
        """Generic crop coefficient for annual row crops (class 1)

        Parameters
        ----------
        fc : ee.Image
            Fraction of cover

        Returns
        -------
        ee.Image

        References
        ----------
        Melton, F., L. Johnson, C. Lund, L. Pierce, A. Michaelis, S. Hiatt,
            A. Guzman, D. Adhikari, A. Purdy, C. Rosevelt, P. Votava, T. Trout,
            B. Temesgen, K. Frame, E. Sheffner, and R. Nemani (2012).
            Satellite Irrigation Management Support with the Terrestrial
            Observation and Prediction System: An Operational Framework for
            Integration of Satellite and Surface Observations to Support
            Improvements in Agricultural Water Resource Management.
            IEEE J. Selected Topics in Applied Earth Observations & Remote Sensing
            5:1709-1721.  [FIG 2b]

        """
        return (
            fc.expression("((b(0) ** 2) * -0.4771) + (1.4047 * b(0)) + 0.15")
            .rename(['kc'])
        )

    def kc_tree(self, fc):
        """General crop coefficient for tree crops (class 3)

        Parameters
        ----------
        fc : ee.Image
            Fraction of cover

        Returns
        -------
        ee.Image

        References
        ----------
        Ayars, J., R. Johnson, C. Phene, T. Trout, D. Clark, and R. Mead, 2003.
            Water use by drip-irrigated late-season peaches.
            Irrigation Science 22:187-194.  DOI 10.1007/s00271-003-0084-4 [EQN 7]

        """
        return fc.multiply(1.48).add(0.007).rename(['kc'])

    def kc_rice(self, fc, ndvi):
        """Crop coefficient for rice crops (class 5)

        Parameters
        ----------
        fc : ee.Image
            Fraction of cover
        ndvi : ee.Image
            Normalized difference vegetation index

        Returns
        -------
        ee.Image

        Notes
        -----
        Kc is computed using the generic annual crop equation (class 1) but
        the Kc is adjusted to 1.05 for low NDVI pixels.

        """
        return self.kc_row_crop(fc).where(ndvi.lte(0.14), 1.05).rename(['kc'])

    def kc_fallow(self, fc, ndvi):
        """Crop coefficient for fallow crops (class 6)

        Parameters
        ----------
        fc : ee.Image
            Fraction of cover
        ndvi : ee.Image
            Normalized difference vegetation index

        Returns
        -------
        ee.Image

        Notes
        -----
        Kc is computed using the generic annual crop equation (class 1) but
        the Kc is set to the Fc for lower NDVI pixels.

        """
        return self.kc_row_crop(fc).where(ndvi.lte(0.35), fc).max(0.01).rename(['kc'])

    def kc_grass_pasture(self, fc, ndvi):
        """Crop coefficient for grass/pasture crops (class 7)

        Parameters
        ----------
        fc : ee.Image
            Fraction of cover
        ndvi : ee.Image
            Normalized difference vegetation index

        Returns
        -------
        ee.Image

        Notes
        -----
        Kc is computed using the generic annual crop equation (class 1) but
        the Kc is set to the Fc for lower NDVI pixels.

        """
        return self.kc_row_crop(fc).where(ndvi.lte(0.35), fc).max(0.01).rename(['kc'])

    def _kcb(self, kd, kc_min=0.15):
        """Basal crop coefficient (Kcb)

        Parameters
        ----------
        kd : ee.Image
            Crop density coefficient
        kc_min : float, optional

        Returns
        -------
        ee.Image

        Notes
        -----
        Compute fr along the line defined by:
            fr = (DOY - ls_start) * ((fr_end - fr_mid) / (ls_stop - ls_start))
                 + fr_mid
        Negated (DOY - ls_start) since DOY is a number and needs to be
        subtracted from an image (images can't be subtracted from numbers).
            fr = (ls_start - DOY) * ((fr_mid - fr_end) / (ls_stop - ls_start))
                 + fr_mid
        fr min/max calls may seem swapped since fr_mid is greater than fr_end

        References
        ----------
        Allen, R., and L. Pereira (2009).  Estimating crop coefficients from
            fraction of ground cover and height.  Irrigation Science 28:17-34.
            DOI 10.1007/s00271-009-0182-z [EQNS 5a, 7a]

        """
        # Reduction factor for adjusting Kcb of tree crops
        fr = (
            self.ls_start.subtract(self.doy)
            .multiply(self.fr_mid.subtract(self.fr_end))
            .divide(self.ls_stop.subtract(self.ls_start))
            .add(self.fr_mid)
            .max(self.fr_end)
            .min(self.fr_mid)
        )

        # Kcb during peak plant growth (near full cover)
        kcb_full = self.h_max.multiply(0.1).add(1).min(1.2).multiply(fr)

        return kd.multiply(kcb_full.subtract(kc_min)).add(kc_min).rename(['kcb'])

    def _kd_row_crop(self, fc):
        """Density coefficient for annual row crops (class 1)

        Parameters
        ----------
        fc : ee.Image
            Fraction of cover

        Returns
        -------
        ee.Image

        Notes
        -----
        TODO: Add a more readable version of the calculation

        References
        ----------
        Allen, R., and L. Pereira (2009).  Estimating crop coefficients from
            fraction of ground cover and height.  Irrigation Science 28:17-34.
            DOI 10.1007/s00271-009-0182-z
            [EQNS 10 (Kd); 7a (Kcb_full) using tree/vine Fr vals from Table 2]

        """
        # First calculation is the fc.divide(0.7).lte(1) case
        return (
            fc.multiply(self.m_l)
            .min(fc.pow(fc.divide(0.7).multiply(self.h_max).add(1).pow(-1)))
            .where(
                fc.divide(0.7).gt(1),
                fc.multiply(self.m_l).min(fc.pow(self.h_max.add(1).pow(-1))))
            .min(1)
            .rename(['kd'])
        )

    def _kd_vine(self, fc):
        """Crop coefficient for vine crops (class 2)

        Parameters
        ----------
        fc : ee.Image
            Fraction of cover

        Returns
        -------
        ee.Image

        References
        ----------
        CGM - I couldn't find this equation in FIG 10 of the reference
        Williams, L. and J. Ayars (2005).  Grapevine water use and the crop
            coefficient are linear functions of the shaded area measured beneath
            the canopy.  Ag. For. Meteor 132:201-211.  [FIG 10]

        """
        return fc.multiply(1.5).min(fc.pow(1 / (1 + 2))).min(1).rename(['kd'])

    def _kd_tree(self, fc):
        """Density coefficient for tree crops (class 3)

        Parameters
        ----------
        fc : ee.Image
            Fraction of cover

        Returns
        -------
        ee.Image

        References
        ----------
        Allen, R., and L. Pereira (2009).  Estimating crop coefficients from
            fraction of ground cover and height.  Irrigation Science 28:17-34.
            DOI 10.1007/s00271-009-0182-z
            [EQNS 10 (Kd); 7a (Kcb_full) using tree/vine Fr vals from Table 2; 5a (Kcb)]

        """
        # First calculation is the fc.gt(0.5) case
        return (
            fc.multiply(self.m_l).min(fc.pow(self.h_max.add(1).pow(-1)))
            .where(fc.lte(0.5), fc.multiply(self.m_l).min(fc.pow(self.h_max.pow(-1))))
            .min(1)
            .rename(['kd'])
        )


def crop_data_image(param_name, crop_type, crop_data, default_value=None):
    """Build a constant ee.Image of crop type data for one parameter

    Parameters
    ----------
    param_name : str
    crop_type : ee.Image
    crop_data : dict
        Imported from data.py
    default_value : float, optional
        The default value to replace values that weren't matched in the remap.
        If default_value is not set or is None, unmatched values are masked.

    Returns
    -------
    ee.Image

    Notes
    -----
    All values are multiplied by an factor (read from the data.py file),
    then divided after the remap in order to return floating point values.

    """
    data_dict = {
        c_type: round(c_data[param_name] * data.int_scalar)
        for c_type, c_data in crop_data.items()
        if param_name in c_data.keys()
    }
    from_list, to_list = zip(*sorted(data_dict.items()))

    # Scale the default value if it is set
    if default_value is not None:
        output = crop_type.remap(from_list, to_list, default_value * data.int_scalar)
    else:
        output = crop_type.remap(from_list, to_list)

    return output.double().divide(data.int_scalar).rename([param_name])



import datetime
# import pprint

import ee
import pandas as pd
import pytest

import openet.sims.interpolate as interpolate
import openet.sims.utils as utils
import openet.sims as sims


def scene_coll(variables, etf=[0.4, 0.4, 0.4], et=[5, 5, 5], ndvi=[0.6, 0.6, 0.6]):
    """Return a generic scene collection to test scene interpolation functions

    Parameters
    ----------
    variables : list
        The variables to return in the collection
    et_fraction : list
    et : list
    ndvi : list

    Returns
    -------
    ee.ImageCollection

    """
    img = (
        ee.Image('LANDSAT/LC08/C02/T1_L2/LC08_044033_20170716')
        .select(['SR_B3']).double().multiply(0)
    )
    mask = img.add(1).updateMask(1).uint8()

    # # The "date" is used for the time band since it needs to be the 0 UTC time
    # date1 = ee.Number(ee.Date.fromYMD(2017, 7, 8).millis())
    # date2 = ee.Number(ee.Date.fromYMD(2017, 7, 16).millis())
    # date3 = ee.Number(ee.Date.fromYMD(2017, 7, 24).millis())

    # The "time" is advanced to match the typical Landsat overpass time
    time1 = ee.Number(ee.Date.fromYMD(2017, 7, 8).advance(18, 'hours').millis())
    time2 = ee.Number(ee.Date.fromYMD(2017, 7, 16).advance(18, 'hours').millis())
    time3 = ee.Number(ee.Date.fromYMD(2017, 7, 24).advance(18, 'hours').millis())

    # TODO: Add code to convert et, et_fraction, and ndvi to lists if they
    #   are set as a single value

    # Don't add mask or time band to scene collection
    # since they are now added in the interpolation calls
    scene_coll = ee.ImageCollection.fromImages([
        ee.Image([img.add(etf[0]), img.add(et[0]), img.add(ndvi[0])])
        .rename(['et_fraction', 'et', 'ndvi'])
        .set({'system:index': 'LE07_044033_20170708', 'system:time_start': time1}),
        ee.Image([img.add(etf[1]), img.add(et[1]), img.add(ndvi[1])])
        .rename(['et_fraction', 'et', 'ndvi'])
        .set({'system:index': 'LC08_044033_20170716', 'system:time_start': time2}),
        ee.Image([img.add(etf[2]), img.add(et[2]), img.add(ndvi[2])])
        .rename(['et_fraction', 'et', 'ndvi'])
        .set({'system:index': 'LE07_044033_20170724', 'system:time_start': time3}),
    ])

    # # Mask and time bands currently get added on to the scene collection
    # #   and images are unscaled just before interpolating in the export tool
    # scene_coll = ee.ImageCollection([
    #     ee.Image([img.add(etf[0]), img.add(et[0]), img.add(ndvi[0]), img.add(date1), mask])
    #     .rename(['et_fraction', 'et', 'ndvi', 'time', 'mask'])
    #     .set({'system:index': 'LE07_044033_20170708', 'system:time_start': time1}),
    #     ee.Image([img.add(etf[1]), img.add(et[1]), img.add(ndvi[1]), img.add(date2), mask])
    #     .rename(['et_fraction', 'et', 'ndvi', 'time', 'mask'])
    #     .set({'system:index': 'LC08_044033_20170716', 'system:time_start': time2}),
    #     ee.Image([img.add(etf[2]), img.add(et[2]), img.add(ndvi[2]), img.add(date3), mask])
    #     .rename(['et_fraction', 'et', 'ndvi', 'time', 'mask'])
    #     .set({'system:index': 'LE07_044033_20170724', 'system:time_start': time3}),
    # ])

    return scene_coll.select(variables)


def test_from_scene_et_fraction_t_interval_daily_values(tol=0.0001):
    output_coll = interpolate.from_scene_et_fraction(
        scene_coll(['et_fraction', 'ndvi'], ndvi=[0.2, 0.4, 0.6]),
        start_date='2017-07-01', end_date='2017-08-01',
        variables=['et', 'et_reference', 'et_fraction', 'ndvi'],
        interp_args={'interp_method': 'linear', 'interp_days': 32},
        model_args={'et_reference_source': 'IDAHO_EPSCOR/GRIDMET',
                    'et_reference_band': 'eto',
                    'et_reference_resample': 'nearest'},
        t_interval='daily',
    )

    TEST_POINT = (-121.5265, 38.7399)
    output = utils.point_coll_value(output_coll, TEST_POINT, scale=30)
    assert abs(output['ndvi']['2017-07-01'] - 0.2) <= tol
    assert abs(output['ndvi']['2017-07-08'] - 0.2) <= tol
    assert abs(output['ndvi']['2017-07-10'] - 0.25) <= tol
    assert abs(output['ndvi']['2017-07-12'] - 0.3) <= tol
    assert abs(output['ndvi']['2017-07-16'] - 0.4) <= tol
    assert abs(output['ndvi']['2017-07-24'] - 0.6) <= tol
    assert abs(output['ndvi']['2017-07-31'] - 0.6) <= tol
    assert abs(output['et_fraction']['2017-07-10'] - 0.4) <= tol
    assert abs(output['et_reference']['2017-07-10'] - 8.0) <= tol
    assert abs(output['et']['2017-07-10'] - (8.0 * 0.4)) <= tol
    assert abs(output['et_fraction']['2017-07-01'] - 0.4) <= tol
    assert abs(output['et_fraction']['2017-07-31'] - 0.4) <= tol
    assert '2017-08-01' not in output['et_fraction'].keys()
    # assert output['count']['2017-07-01'] == 3


def test_from_scene_et_fraction_t_interval_monthly_values(tol=0.0001):
    output_coll = interpolate.from_scene_et_fraction(
        scene_coll(['et_fraction', 'ndvi']),
        start_date='2017-07-01', end_date='2017-08-01',
        variables=['et', 'et_reference', 'et_fraction', 'ndvi', 'count'],
        interp_args={'interp_method': 'linear', 'interp_days': 32},
        model_args={'et_reference_source': 'IDAHO_EPSCOR/GRIDMET',
                    'et_reference_band': 'eto',
                    'et_reference_resample': 'nearest'},
        t_interval='monthly',
    )

    TEST_POINT = (-121.5265, 38.7399)
    output = utils.point_coll_value(output_coll, TEST_POINT, scale=30)
    assert abs(output['ndvi']['2017-07-01'] - 0.6) <= tol
    assert abs(output['et_fraction']['2017-07-01'] - 0.4) <= tol
    assert abs(output['et_reference']['2017-07-01'] - 236.5) <= tol
    assert abs(output['et']['2017-07-01'] - (236.5 * 0.4)) <= tol
    assert output['count']['2017-07-01'] == 3


def test_from_scene_et_fraction_t_interval_custom_values(tol=0.0001):
    output_coll = interpolate.from_scene_et_fraction(
        scene_coll(['et_fraction', 'ndvi']),
        start_date='2017-07-01', end_date='2017-08-01',
        variables=['et', 'et_reference', 'et_fraction', 'ndvi', 'count'],
        interp_args={'interp_method': 'linear', 'interp_days': 32},
        model_args={'et_reference_source': 'IDAHO_EPSCOR/GRIDMET',
                    'et_reference_band': 'eto',
                    'et_reference_resample': 'nearest'},
        t_interval='custom',
    )

    TEST_POINT = (-121.5265, 38.7399)
    output = utils.point_coll_value(output_coll, TEST_POINT, scale=30)
    assert abs(output['ndvi']['2017-07-01'] - 0.6) <= tol
    assert abs(output['et_fraction']['2017-07-01'] - 0.4) <= tol
    assert abs(output['et_reference']['2017-07-01'] - 236.5) <= tol
    assert abs(output['et']['2017-07-01'] - (236.5 * 0.4)) <= tol
    assert output['count']['2017-07-01'] == 3


def test_from_scene_et_fraction_t_interval_custom_daily_count(tol=0.0001):
    output_coll = interpolate.from_scene_et_fraction(
        scene_coll(['et_fraction', 'ndvi']),
        start_date='2017-07-01', end_date='2017-08-01',
        variables=['et_fraction', 'daily_count'],
        interp_args={'interp_method': 'linear', 'interp_days': 32},
        model_args={'et_reference_source': 'IDAHO_EPSCOR/GRIDMET',
                    'et_reference_band': 'eto',
                    'et_reference_resample': 'nearest'},
        t_interval='custom',
    )

    TEST_POINT = (-121.5265, 38.7399)
    output = utils.point_coll_value(output_coll, TEST_POINT, scale=30)
    assert abs(output['et_fraction']['2017-07-01'] - 0.4) <= tol
    assert output['daily_count']['2017-07-01'] == 31


def test_from_scene_et_fraction_t_interval_monthly_et_reference_factor(tol=0.0001):
    output_coll = interpolate.from_scene_et_fraction(
        scene_coll(['et_fraction', 'ndvi']),
        start_date='2017-07-01', end_date='2017-08-01',
        variables=['et', 'et_reference', 'et_fraction', 'ndvi', 'count'],
        interp_args={'interp_method': 'linear', 'interp_days': 32},
        model_args={'et_reference_source': 'IDAHO_EPSCOR/GRIDMET',
                    'et_reference_band': 'eto',
                    'et_reference_factor': 0.5,
                    'et_reference_resample': 'nearest'},
        t_interval='monthly',
    )

    TEST_POINT = (-121.5265, 38.7399)
    output = utils.point_coll_value(output_coll, TEST_POINT, scale=30)
    assert abs(output['ndvi']['2017-07-01'] - 0.6) <= tol
    assert abs(output['et_fraction']['2017-07-01'] - 0.4) <= tol
    assert abs(output['et_reference']['2017-07-01'] - 236.5 * 0.5) <= tol
    assert abs(output['et']['2017-07-01'] - (236.5 * 0.5 * 0.4)) <= tol
    assert output['count']['2017-07-01'] == 3


def test_from_scene_et_fraction_t_interval_monthly_et_reference_resample(tol=0.0001):
    output_coll = interpolate.from_scene_et_fraction(
        scene_coll(['et_fraction', 'ndvi']),
        start_date='2017-07-01', end_date='2017-08-01',
        variables=['et', 'et_reference', 'et_fraction', 'ndvi', 'count'],
        interp_args={'interp_method': 'linear', 'interp_days': 32},
        model_args={'et_reference_source': 'IDAHO_EPSCOR/GRIDMET',
                    'et_reference_band': 'eto',
                    'et_reference_resample': 'bilinear'},
        t_interval='monthly',
    )

    TEST_POINT = (-121.5265, 38.7399)
    output = utils.point_coll_value(output_coll, TEST_POINT, scale=30)
    assert abs(output['ndvi']['2017-07-01'] - 0.6) <= tol
    assert abs(output['et_fraction']['2017-07-01'] - 0.4) <= tol
    # CGM - ETo (and ET) test values are slightly different with bilinear resampling
    #   but ET fraction should be the same
    assert abs(output['et_reference']['2017-07-01'] - 236.05609) <= tol
    assert abs(output['et']['2017-07-01'] - (236.05609 * 0.4)) <= tol
    # assert abs(output['et_reference']['2017-07-01'] - 236.5) <= tol
    # assert abs(output['et']['2017-07-01'] - (236.5 * 0.4)) <= tol
    assert output['count']['2017-07-01'] == 3


def test_from_scene_et_fraction_t_interval_monthly_interp_args_et_reference(tol=0.0001):
    # Check that the et_reference parameters can be set through the interp_args
    output_coll = interpolate.from_scene_et_fraction(
        scene_coll(['et_fraction', 'ndvi']),
        start_date='2017-07-01', end_date='2017-08-01',
        variables=['et', 'et_reference', 'et_fraction', 'ndvi', 'count'],
        interp_args={'interp_method': 'linear', 'interp_days': 32,
                     'et_reference_source': 'IDAHO_EPSCOR/GRIDMET',
                     'et_reference_band': 'eto',
                     'et_reference_resample': 'nearest'},
        model_args={},
        t_interval='monthly',
    )

    TEST_POINT = (-121.5265, 38.7399)
    output = utils.point_coll_value(output_coll, TEST_POINT, scale=30)
    assert abs(output['ndvi']['2017-07-01'] - 0.6) <= tol
    assert abs(output['et_fraction']['2017-07-01'] - 0.4) <= tol
    assert abs(output['et_reference']['2017-07-01'] - 236.5) <= tol
    assert abs(output['et']['2017-07-01'] - (236.5 * 0.4)) <= tol
    assert output['count']['2017-07-01'] == 3


def test_from_scene_et_fraction_t_interval_bad_value():
    # Function should raise a ValueError if t_interval is not supported
    with pytest.raises(ValueError):
        interpolate.from_scene_et_fraction(
            scene_coll(['et']),
            start_date='2017-07-01', end_date='2017-08-01', variables=['et'],
            interp_args={'interp_method': 'linear', 'interp_days': 32},
            model_args={'et_reference_source': 'IDAHO_EPSCOR/GRIDMET',
                        'et_reference_band': 'etr',
                        'et_reference_factor': 0.5,
                        'et_reference_resample': 'nearest'},
            t_interval='deadbeef',
        )


def test_from_scene_et_fraction_t_interval_no_value():
    # Function should raise an Exception if t_interval is not set
    with pytest.raises(TypeError):
        interpolate.from_scene_et_fraction(
            scene_coll(['et']),
            start_date='2017-07-01', end_date='2017-08-01',
            variables=['et', 'et_reference', 'et_fraction', 'count'],
            interp_args={'interp_method': 'linear', 'interp_days': 32},
            model_args={'et_reference_source': 'IDAHO_EPSCOR/GRIDMET',
                        'et_reference_band': 'etr',
                        'et_reference_factor': 0.5,
                        'et_reference_resample': 'nearest'},
        )


def test_soil_evaporation_landsat(tol=0.001):
    TEST_POINT = (-120.201, 36.1696)
    et_reference_source = 'IDAHO_EPSCOR/GRIDMET'
    et_reference_band = 'eto'
    start_date = '2018-02-16'
    end_date = '2018-03-08'

    # Goal is to have the following two images in the test collection
    # ['LC08_043035_20180218', 'LC08_043035_20180306']
    landsat_coll = (
        ee.ImageCollection('LANDSAT/LC08/C02/T1_L2')
        .filterDate(start_date, end_date)
        .filterMetadata('WRS_PATH', 'equals', 43)
        .filterMetadata('WRS_ROW', 'equals', 35)
        # .filterBounds(ee.Geometry.Point(TEST_POINT))
    )

    def make_et_frac(img):
        # if 'C02' in landsat_coll_id:
        mask_img = img.select(['SR_B3']).double().multiply(0).rename(['mask'])
        time_img = mask_img.add(ee.Number(img.get('system:time_start'))).rename(['time'])
        et_img = sims.Image.from_landsat_c2_sr(
            img,
            et_reference_source=et_reference_source,
            et_reference_band=et_reference_band,
            cloudmask_args={'cloud_score_flag': False, 'filter_flag': False},
        ).calculate(['ndvi', 'et_reference', 'et_fraction', 'et'])
        return et_img.addBands([time_img])

    test_coll = landsat_coll.map(make_et_frac)

    # # Build the collection manually
    # img_a = ee.Image('LANDSAT/LC08/C02/T1_L2/LC08_043035_20180218')
    # img_b = ee.Image('LANDSAT/LC08/C02/T1_L2/LC08_043035_20180306')
    # time_a = (
    #     img_a.select(['SR_B3'], ['time'])
    #     .double().multiply(0).add(ee.Number(img_a.get('system:time_start')))
    # )
    # time_b = (
    #     img_b.select(['SR_B3'], ['time'])
    #     .double().multiply(0).add(ee.Number(img_b.get('system:time_start')))
    # )
    # img_a = sims.Image.from_landsat_c2_sr(
    #         img_a,
    #         et_reference_source=et_reference_source,
    #         et_reference_band=et_reference_band,
    #         cloudmask_args={'cloud_score_flag': False, 'filter_flag': False},
    #     ).calculate(['ndvi', 'et_reference', 'et_fraction', 'et'])
    # img_b = sims.Image.from_landsat_c2_sr(
    #         img_b,
    #         et_reference_source=et_reference_source,
    #         et_reference_band=et_reference_band,
    #         cloudmask_args={'cloud_score_flag': False, 'filter_flag': False},
    #     ).calculate(['ndvi', 'et_reference', 'et_fraction', 'et'])
    # test_coll = ee.ImageCollection([img_a.addBands([time_a]), img_b.addBands([time_b])])

    normal_coll = interpolate.from_scene_et_fraction(
        test_coll,
        start_date=start_date,
        end_date=end_date,
        variables=['et_reference', 'et_fraction', 'et'],
        interp_args={'interp_method': 'linear', 'interp_days': 32, 'use_joins': True},
        model_args={'et_reference_source': 'IDAHO_EPSCOR/GRIDMET',
                    'et_reference_band': 'eto',
                    'et_reference_resample': 'nearest'},
        t_interval='daily',
    )
    normal = utils.point_coll_value(normal_coll, TEST_POINT, scale=30)

    wb_coll = interpolate.from_scene_et_fraction(
        test_coll,
        start_date=start_date,
        end_date=end_date,
        variables=['et_reference', 'et_fraction', 'ke', 'et', 'ndvi'],
        interp_args={'interp_method': 'linear', 'interp_days': 32,
                     'estimate_soil_evaporation': True, 'use_joins': True},
        model_args={'et_reference_source': 'IDAHO_EPSCOR/GRIDMET',
                    'et_reference_band': 'eto',
                    'et_reference_resample': 'nearest'},
        t_interval='daily',
    )
    wb = utils.point_coll_value(wb_coll, TEST_POINT, scale=30)

    for date in normal['et'].keys():
        # check that ET with soil evap >= ET without soil evap
        assert wb['et'][date] >= normal['et'][date]


# Global constants for soil evap tests
TEST_POINT = (-120.201, 36.1696)
start_date = '2018-02-10'
end_date = '2018-03-14'
comp_df = pd.read_csv('openet/sims/tests/ee_wb_valid.csv')
comp_df['et_fraction'] = comp_df['etc'] / comp_df['eto']


@pytest.fixture
def synth_test_imgs():
    landsat_coll = (
        ee.ImageCollection('LANDSAT/LC08/C02/T1_L2')
        .filterDate(start_date, end_date)
        .filterBounds(ee.Geometry.Point(TEST_POINT))
    )
    mask = landsat_coll.first().select(['SR_B3']).double().multiply(0)

    test_imgs = []
    for index, row in comp_df.iterrows():
        # The time band needs to be 0 UTC for the interpolation to work correctly
        date = ee.Date.fromYMD(2018, 1, 1).advance(row.doy-1, 'days')
        time = date.advance(16, 'hours')
        test_img = (
            ee.Image([mask.add(ee.Number(date.millis())),
                      mask.add(row.ndvi_interp),
                      mask.add(row.kc), mask.add(row.eto)])
            .rename(['time', 'ndvi', 'et_fraction', 'et_reference'])
            .set({'system:time_start': ee.Number(time.millis()),
                  'system:index': date.format('yyyyMMdd')})
        )
        test_imgs.append(test_img)

    return ee.ImageCollection(test_imgs)


@pytest.fixture
def synth_precip_imgs():
    mask = (
        ee.ImageCollection('IDAHO_EPSCOR/GRIDMET')
        .filterDate(start_date, end_date)
        .first().select(['pr']).multiply(0)
    )

    precip_imgs = []
    for index, row in comp_df.iterrows():
        # Add 6 hours to mimic GRIDMET start time
        date = ee.Date.fromYMD(2018, 1, 1).advance(row.doy-1, 'days').advance(6, 'hours')
        precip_img = (
            mask.add(row.pr).rename(['pr'])
            .set({'system:time_start': ee.Number(date.millis()),
                  'system:index': date.format('yyyyMMdd')})
        )
        precip_imgs.append(precip_img)

    # Precipitation collection needs extra images at the end since the
    #   daily_ke builds a "next_precip" image in the iterate block
    date = ee.Date.fromYMD(2018, 1, 1).advance(comp_df.iloc[-1].doy, 'days')\
        .advance(6, 'hours')
    precip_imgs.append(mask.add(0).rename(['pr']).set({
        'system:time_start': ee.Number(date.millis()),
        'system:index': date.format('yyyyMMdd')}))

    return ee.ImageCollection(precip_imgs)


def test_daily_ke(synth_test_imgs, synth_precip_imgs):
    """"""
    evap_imgs = interpolate.daily_ke(
        synth_test_imgs,
        # CGM - model_args isn't used by daily_ke function and could be removed
        model_args={},
        precip_source=synth_precip_imgs, precip_band='pr',
        # precip_source='IDAHO_EPSCOR/GRIDMET', precip_band='pr',
        fc_source='projects/eeflux/soils/gsmsoil_mu_a_fc_10cm_albers_100',
        fc_band='b1',
        wp_source='projects/eeflux/soils/gsmsoil_mu_a_wp_10cm_albers_100',
        wp_band='b1',
    )

    base_ts = utils.point_coll_value(synth_test_imgs, TEST_POINT, scale=30)
    base_df = pd.DataFrame(base_ts).reset_index()
    base_df['et'] = base_df['et_fraction'] * base_df['et_reference']

    evap_ts = utils.point_coll_value(evap_imgs, TEST_POINT, scale=30)
    evap_df = pd.DataFrame(evap_ts).reset_index()
    evap_df['et'] = evap_df['et_fraction'] * evap_df['et_reference']

    # Iterate through time series, stop one before end to avoid out of bounds
    # in "check next day state variable" tests
    for i in range(evap_df.shape[0]-1):
        # Check that soil evap only increase et fraction
        assert base_df.loc[i, 'et_fraction'] <= evap_df.loc[i, 'et_fraction']

        # Check that etc strictly greater than etcb if it rained and
        # kcb isn't maxed out
        if evap_df.loc[i, 'precip'] > 0 and base_df.loc[i, 'et_fraction'] < 1.15:
            assert base_df.loc[i+1, 'et_fraction'] < evap_df.loc[i+1, 'et_fraction']

        # Check evaporation reduction coefficients
        # Should be nonzero next day when depletion > REW
        if evap_df.loc[i, 'de'] > evap_df.de_rew.max():
            assert evap_df.loc[i+1, 'kr'] < 1

        # should be one next day when depletion is less than REW
        if evap_df.loc[i, 'de'] < evap_df.de_rew.max():
            assert evap_df.loc[i+1, 'kr'] == 1

        # should be zero next day when fully depleted
        if evap_df.loc[i, 'de'] == evap_df.de.max():
            assert evap_df.loc[i+1, 'kr'] == 0


def test_soil_evap_fails_without_ndvi(synth_test_imgs):
    """Test that daily_ke raises exception if `ndvi` band not present"""
    try:
        interpolate.from_scene_et_fraction(
            synth_test_imgs,
            start_date=start_date,
            end_date=end_date,
            variables=['et_reference', 'et_fraction', 'ke', 'et', 'precip'],
            interp_args={'interp_method': 'linear', 'interp_days': 10,
                         'estimate_soil_evaporation': True},
            model_args={'et_reference_source': 'provided',
                        'et_reference_band': 'eto',
                        'et_reference_resample': 'nearest'},
            t_interval='daily',
        )
        # if from_scene_et_fraction doesn't raise, assert False
        assert False
    except Exception:
        pass


def test_soil_evaporation_synthetic(synth_test_imgs, synth_precip_imgs, tol=0.001):
    """Test that setting 'estimate_soil_evaporation' flag runs SWB"""
    normal_coll = interpolate.from_scene_et_fraction(
        scene_coll=synth_test_imgs,
        start_date=start_date,
        end_date=end_date,
        variables=['et_reference', 'et_fraction', 'et', 'ndvi'],
        interp_args={'interp_method': 'linear', 'interp_days': 10,
                     'estimate_soil_evaporation': False},
        model_args={'et_reference_source': 'provided',
                    'et_reference_band': 'eto',
                    'et_reference_resample': 'nearest'},
        t_interval='daily',
    )

    wb_coll = interpolate.from_scene_et_fraction(
        scene_coll=synth_test_imgs,
        start_date=start_date,
        end_date=end_date,
        variables=['et_reference', 'et_fraction', 'ke', 'et', 'ndvi', 'precip'],
        interp_args={'interp_method': 'linear', 'interp_days': 10,
                     'estimate_soil_evaporation': True,
                     'precip_source': synth_precip_imgs,
                     'precip_band': 'pr'},
        model_args={'et_reference_source': 'provided',
                    'et_reference_band': 'eto',
                    'et_reference_resample': 'nearest'},
        t_interval='daily',
    )

    # synth = utils.point_coll_value(synth_test_imgs, TEST_POINT, scale=30)
    normal = utils.point_coll_value(normal_coll, TEST_POINT, scale=30)
    wb = utils.point_coll_value(wb_coll, TEST_POINT, scale=30)

    # check that wb ET >= regular SIMS ET
    for date in normal['et'].keys():
        assert wb['et'][date] >= normal['et'][date]

    def get_doy(dt_str):
        return datetime.datetime.strptime(dt_str, '%Y-%m-%d').timetuple().tm_yday

    wb_df = pd.DataFrame(wb)
    wb_df = wb_df.reset_index()
    wb_df['doy'] = wb_df['index'].apply(get_doy)

    for i in range(46, 52):
        assert abs(wb_df[wb_df.doy==i]['et'].iloc[0] -
                   comp_df[comp_df.doy==i]['etc'].iloc[0]) < tol
